"""Test smol.moca.utils.math_utils."""

from collections import Counter
from fractions import Fraction

import numpy as np
import numpy.testing as npt
import polytope as pc
import pytest

from smol.utils.math import (
    NUM_TOL,
    choose_section_from_partition,
    compute_snf,
    connectivity,
    count_row_matches,
    flip_size,
    flip_weights_mask,
    get_ergodic_vectors,
    get_natural_centroid,
    get_natural_solutions,
    get_nonneg_float_vertices,
    get_one_dim_solutions,
    get_optimal_basis,
    integerize_multiple,
    integerize_vector,
    rationalize_number,
    solve_diophantines,
    yield_hermite_normal_forms,
)
from tests.utils import assert_table_set_equal

a1 = [[1, 3, 4, -3, -2], [1, 1, 1, 0, 0], [0, 0, 0, 1, 1]]
b1 = [0, 1, 1]  # LMTPO, prim

a2 = [
    [1, 2, 3, 4, -2, -1],
    [1, 1, 1, 1, 0, 0],
    [0, 0, 0, 0, 1, 1],
    [0, 1, -1, 0, 0, 0],
    [0, 0, 1, -1, 0, 0],
]
b2 = [0, 1, 1, 0, 0]  # LNMTOF, prim

a3 = [
    [1, 2, 3, 4, 5, 0, -2, -1, -1],
    [1, 0, 0, 0, 0, 1, 0, 0, 0],
    [0, 1, 1, 1, 0, 0, 0, 0, 0],
    [0, 0, 0, 0, 1, 0, 0, 0, 0],
    [0, 0, 0, 0, 0, 0, 1, 1, 0],
    [0, 0, 0, 0, 0, 0, 0, 0, 1],
]
b3 = [0, 3, 2, 1, 5, 1]  # L M(234) Nb Vac O(-2-1) F, topotactic, prim.

a4 = [[0, 0, 0, 0], [1, 1, 0, 0], [0, 0, 1, 1]]
b4 = [0, 1, 2]  # Some random charge neutral alloy of 2 sub-lattices, prim.


all_ab = [(a1, b1), (a2, b2), (a3, b3), (a4, b4)]


@pytest.fixture(scope="module", params=all_ab)
def a(request):
    # The matrix A in snf.
    return np.array(request.param[0], dtype=int)


@pytest.fixture(scope="module", params=all_ab)
def ab(request):
    # The matrix A in snf.
    return (
        np.array(request.param[0], dtype=int),
        np.array(request.param[1], dtype=int),
    )


def test_rational():
    for _ in range(500):
        rand_num, rand_den = np.random.randint(low=-1000, high=1000, size=2).tolist()
        g = np.gcd(rand_num, rand_den)
        rand_num = rand_num // g
        rand_den = rand_den // g
        if rand_den < 0:
            rand_den = -rand_den
        if rand_den == 0:
            rand_den = 1
        perturbation = np.random.rand() * NUM_TOL * 0.5
        x = float(rand_num) / rand_den + perturbation
        num, den = rationalize_number(x, max_denominator=1000, dtol=NUM_TOL)
        if rand_num != 0:
            assert num == rand_num
            assert den == rand_den

        rand_num, rand_den = 0, np.random.randint(1000)
        if rand_den == 0:
            rand_den = 1
        perturbation = np.random.rand() * NUM_TOL * 0.5
        x = float(rand_num) / rand_den + perturbation
        num, den = rationalize_number(x, max_denominator=1000, dtol=NUM_TOL)
        assert num == 0
        assert den == 1

    # Test bad cases.
    for _ in range(500):
        rand_num, rand_den = np.random.randint(low=-1000, high=1000, size=2).tolist()
        if rand_den < 0:
            rand_den = -rand_den
        if rand_den == 0:
            rand_den = 1
        perturbation = np.random.rand() * NUM_TOL * 10
        x = float(rand_num) / rand_den + perturbation
        f = Fraction.from_float(x)
        f2 = Fraction.from_float(x).limit_denominator(1000)
        num = f.numerator
        den = f.denominator
        num2 = f2.numerator
        den2 = f2.denominator
        if abs(num2 / den2 - num / den) > NUM_TOL:
            with pytest.raises(ValueError):
                _, _ = rationalize_number(x, max_denominator=1000, dtol=NUM_TOL)


def test_integerize():
    for _ in range(100):
        a = np.random.randint(low=-10000, high=10000, size=100)
        aa = a.reshape((10, 10))
        den1, den2 = np.random.randint(1000, size=2).tolist()
        if den1 == 0:
            den1 = 1
        if den2 == 0:
            den2 = 1
        g1 = np.gcd(np.gcd.reduce(a), den1)
        g2 = np.gcd(np.gcd.reduce(a), den2)
        a = a // g1
        den1 = den1 // g1
        aa = aa // g2
        den2 = den2 // g2

        pert = np.random.rand(100) * NUM_TOL * 0.5
        a_t = a / den1 + pert
        aa_t = aa / den2 + pert.reshape((10, 10))

        a_r, den1_r = integerize_vector(a_t, max_denominator=1000, dtol=NUM_TOL)
        aa_r, den2_r = integerize_multiple(aa_t, max_denominator=1000, dtol=NUM_TOL)

        npt.assert_array_equal(a_r, a)
        npt.assert_array_equal(aa_r, aa)
        assert den1 == den1_r
        assert den2 == den2_r

    a = np.zeros(100, dtype=int)
    a_r, den_r = integerize_vector(a, max_denominator=1000, dtol=NUM_TOL)
    npt.assert_array_equal(a_r, a)
    assert den_r == 1


def test_snf(a):
    rank = np.linalg.matrix_rank(a)
    s, m, t = compute_snf(a)
    assert m.shape == a.shape
    npt.assert_array_equal(m, s @ a @ t)
    m_diag = m.diagonal().copy()
    m_diag = m_diag[m_diag != 0]
    assert len(m_diag) == rank
    npt.assert_array_equal(*np.nonzero(m))

    n, d = a.shape
    for _ in range(100):
        x = np.random.randint(low=-10, high=10, size=(d, d - rank))
        npt.assert_array_equal(a @ (t[:, rank:] @ x.T), 0)


def test_snf_specific():
    # Specific test from wikipedia
    a = np.array([[2, 4, 4], [-6, 6, 12], [10, 4, 16]])
    s, m, t = compute_snf(a)
    npt.assert_array_equal(m.diagonal(), [2, 2, 156])
    npt.assert_array_equal(m, s @ a @ t)

    # Zeros
    a = np.zeros((50, 100), dtype=int)
    s, m, t = compute_snf(a)
    npt.assert_array_equal(m, 0)
    assert_table_set_equal(s, np.eye(50, dtype=int))
    assert_table_set_equal(t, np.eye(100, dtype=int))


def test_snf_rand():
    for _ in range(10):
        a = np.random.randint(low=-8, high=8, size=(10, 20))
        rank = np.linalg.matrix_rank(a)
        s, m, t = compute_snf(a)
        assert m.shape == a.shape
        npt.assert_array_equal(m, s @ a @ t)
        m_diag = m.diagonal().copy()
        m_diag = m_diag[m_diag != 0]
        assert len(m_diag) == rank
        npt.assert_array_equal(*np.nonzero(m))

        n, d = a.shape
        for _ in range(10):
            x = np.random.randint(low=-100, high=100, size=(d, d - rank))
            npt.assert_array_equal(a @ (t[:, rank:] @ x.T), 0)


# This can cause some issues when ran on GitHub tests. Not fully clear what happened.
def test_solve_diop_rand():
    fail_counts = 0
    for _ in range(500):
        a = np.random.randint(low=-8, high=8, size=(10, 20))
        rank = np.linalg.matrix_rank(a)
        m, d = a.shape
        nn = np.random.randint(low=0, high=10, size=20)
        b = a @ nn
        # TODO: maybe find what causes numerical issues someday.
        try:  # Overwrite numerical issues.
            n0, vs = solve_diophantines(a, b)
            assert vs.shape == (d - rank, d)
            x = np.random.randint(low=-100, high=100, size=(100, d - rank))
            ns = x @ vs + n0
            npt.assert_array_equal(a @ ns.T - b[:, None], 0)
        except ValueError:
            fail_counts += 1
    assert fail_counts < 500


def test_solve_diop_specific(ab):
    a, b = ab
    rank = np.linalg.matrix_rank(a)
    m, d = a.shape
    n0, vs = solve_diophantines(a, b)
    assert vs.shape == (d - rank, d)
    x = np.random.randint(low=-100, high=100, size=(100, d - rank))
    ns = x @ vs + n0
    npt.assert_array_equal(a @ ns.T - b[:, None], 0)


def test_float_verts_specific(ab):
    # Can not do fully random test because of some weird
    # Numerical issues.
    a, b = ab
    verts = get_nonneg_float_vertices(a, b)
    n, d = a.shape
    assert verts.shape[1] == d
    assert np.all(np.any(np.isclose(verts, 0, atol=NUM_TOL), axis=-1))
    npt.assert_almost_equal(a @ verts.T - b[:, None], 0, decimal=6)
    assert np.all(verts >= 0 - NUM_TOL)
    # Also checked numbers, they are reasonable.


def test_centroid_specific(ab):
    # Can not do fully random test because random polytopes might not
    # be full dimensional, neither might it be bounded.
    a, b = ab
    b = b * 10

    n0, vs = solve_diophantines(a, b)
    x_cent = get_natural_centroid(n0, vs, 10)
    n_cent = x_cent @ vs + n0
    npt.assert_array_equal(a @ n_cent, b)
    assert np.all(n_cent >= 0)
    # Can not test for optimality.


def test_centroid_conditioned():
    a = np.array([[1, 3, 4, -3, -2], [1, 1, 1, 0, 0], [0, 0, 0, 1, 1]])
    b = np.array([0, 1, 1]) * 12  # LMTPO, sc_size = 12

    a_leq = np.array(
        [[0, 1, 0, 0, 0], [0, 0, 0, 1, 0], [0, -1, 0, 0, 0], [0, 0, 0, -1, 0]]
    )
    b_leq = np.array([5 / 6, 5 / 6, -1 / 6, -1 / 6])

    n0, vs = solve_diophantines(a, b)
    poly = pc.Polytope(-vs.transpose(), n0)
    centroid = np.average(pc.extreme(poly), axis=0)
    x_cent = get_natural_centroid(n0, vs, 12)
    x_cent2 = get_natural_centroid(n0, vs, 12, a_leq, b_leq)
    assert np.isclose(
        np.linalg.norm(x_cent - centroid), np.linalg.norm(x_cent2 - centroid)
    )
    # Will allow some degeneracy.

    # When setting 4 / 6 in ECOS_BB, this test can not work. In guorbi 4/6 works.
    # Very strange indeed.
    b_leq2 = np.array([5 / 6, 5 / 6, -1 / 6, -3.5 / 6])  # In per prim unit.
    x_cent3 = get_natural_centroid(n0, vs, 12, a_leq, b_leq2)
    assert np.linalg.norm(x_cent - centroid) < np.linalg.norm(x_cent3 - centroid)


def test_one_dim_solution():
    for _ in range(10):
        n = np.random.randint(low=0, high=1000, size=20)
        v = np.random.randint(low=-1000, high=1000, size=20)
        x0 = np.random.randint(low=-10, high=10)
        n0 = n - x0 * v
        # Now this is going to have at least one solution.

        xs = get_one_dim_solutions(n0, v)
        assert len(xs) > 0
        assert np.all(np.outer(xs, v) + n0 >= 0)
        npt.assert_array_equal(xs, np.arange(min(xs), max(xs) + 1, dtype=int))
        # Check that the bound is tightly filled.
        xs_oob = np.array([min(xs) - 1, max(xs) + 1], dtype=int)
        assert np.all(np.any(np.outer(xs_oob, v) + n0 < 0, axis=-1))

    # Test 3 bad cases with no solution.
    # Case 1: when bounds create an empty set.
    v = np.array([-2, 1, -1, 1, -1, 1])
    n0 = np.array([1, 1, 0, 2, 0, -1])
    xs = get_one_dim_solutions(n0, v)
    assert len(xs) == 0
    assert len(xs.shape) == 1
    # Case 2: actually a special case 1, when a v=0 while n0<0.
    n0 = np.array([1, -1, 0])
    v = np.array([0, 1, -5])
    xs = get_one_dim_solutions(n0, v)
    assert len(xs) == 0
    # Counter example of case 2, and unbounded.
    n0 = np.array([1, -1, 0])
    v = np.array([0, 1, 5])
    with pytest.raises(ValueError):
        _ = get_one_dim_solutions(n0, v)

    # Case 3: when bounds do not include an integer.
    v = np.array([1, 1, -2, 3])
    n0 = np.array([1, 2, 1, -1])
    xs = get_one_dim_solutions(n0, v)
    assert len(xs) == 0


def test_natural_solutions_specific():
    # Since random matrices are usually not bounded, we will not test random.
    # Test specific cases.
    a = np.array([[1, 3, 4, -3, -2], [1, 1, 1, 0, 0], [0, 0, 0, 1, 1]])
    b = np.array([0, 6, 6])
    n0, vs = solve_diophantines(a, b)
    xs = get_natural_solutions(n0, vs)
    ns = xs @ vs + n0
    ns = np.array(sorted(ns.tolist()), dtype=int)
    # Manually obtained solutions.
    ns_std = np.array(
        [
            [4, 0, 2, 0, 6],
            [3, 3, 0, 0, 6],
            [3, 2, 1, 1, 5],
            [3, 1, 2, 2, 4],
            [2, 4, 0, 2, 4],
            [3, 0, 3, 3, 3],
            [2, 3, 1, 3, 3],
            [2, 2, 2, 4, 2],
            [1, 5, 0, 4, 2],
            [2, 1, 3, 5, 1],
            [1, 4, 1, 5, 1],
            [2, 0, 4, 6, 0],
            [1, 3, 2, 6, 0],
            [0, 6, 0, 6, 0],
        ],
        dtype=int,
    )
    ns_std = np.array(sorted(ns_std.tolist()), dtype=int)
    npt.assert_array_equal(ns, ns_std)

    a = np.array([[1, 3, 4, -3, -2], [1, 1, 1, 0, 0], [0, 0, 0, 1, 1]])
    b = np.array([0, 6, 6]) * 2
    n0, vs = solve_diophantines(a, b)
    xs = get_natural_solutions(n0, vs, step=2)  # Test a step.
    ns = xs @ vs + n0
    ns = np.array(sorted(ns.tolist()), dtype=int)
    # Scaling test.
    ns_std = (
        np.array(
            [
                [4, 0, 2, 0, 6],
                [3, 3, 0, 0, 6],
                [3, 2, 1, 1, 5],
                [3, 1, 2, 2, 4],
                [2, 4, 0, 2, 4],
                [3, 0, 3, 3, 3],
                [2, 3, 1, 3, 3],
                [2, 2, 2, 4, 2],
                [1, 5, 0, 4, 2],
                [2, 1, 3, 5, 1],
                [1, 4, 1, 5, 1],
                [2, 0, 4, 6, 0],
                [1, 3, 2, 6, 0],
                [0, 6, 0, 6, 0],
            ],
            dtype=int,
        )
        * 2
    )
    ns_std = np.array(sorted(ns_std.tolist()), dtype=int)
    npt.assert_array_equal(ns, ns_std)

    a = np.array([[1, 3, 4, -2, -1], [1, 1, 1, 0, 0], [0, 0, 0, 1, 1]])
    b = np.array([0, 6, 6])
    n0, vs = solve_diophantines(a, b)
    xs = get_natural_solutions(n0, vs)
    ns = xs @ vs + n0
    ns = np.array(sorted(ns.tolist()), dtype=int)
    # Manually obtained solutions.
    ns_std = np.array(
        [
            [4, 0, 2, 6, 0],
            [3, 3, 0, 6, 0],
            [4, 1, 1, 5, 1],
            [4, 2, 0, 4, 2],
            [5, 0, 1, 3, 3],
            [5, 1, 0, 2, 4],
            [6, 0, 0, 0, 6],
        ],
        dtype=int,
    )
    ns_std = np.array(sorted(ns_std.tolist()), dtype=int)
    npt.assert_array_equal(ns, ns_std)

    a = np.array(
        [
            [1, 2, 3, 4, -2, -1],
            [1, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 1, 1],
            [0, 1, -1, 0, 0, 0],
            [0, 0, 1, -1, 0, 0],
        ]
    )
    b = np.array([0, 12, 12, 0, 0])
    n0, vs = solve_diophantines(a, b)
    xs = get_natural_solutions(n0, vs)
    ns = xs @ vs + n0
    ns = np.array(sorted(ns.tolist()), dtype=int)
    # Manually obtained solutions.
    ns_std = np.array(
        [[12, 0, 0, 0, 0, 12], [9, 1, 1, 1, 6, 6], [6, 2, 2, 2, 12, 0]], dtype=int
    )
    ns_std = np.array(sorted(ns_std.tolist()), dtype=int)
    npt.assert_array_equal(ns, ns_std)

    a = np.array(
        [
            [1, 2, 3, 4, 5, 0, -2, -1, -1],
            [1, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 1, 1, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1],
        ]
    )
    b = np.array([0, 3, 2, 1, 5, 1])
    n0, vs = solve_diophantines(a, b)
    xs = get_natural_solutions(n0, vs)
    ns = xs @ vs + n0
    ns = np.array(sorted(ns.tolist()), dtype=int)
    # Manually obtained solutions.
    ns_std = np.array(
        [
            [2, 2, 0, 0, 1, 1, 5, 0, 1],
            [1, 1, 1, 0, 1, 2, 5, 0, 1],
            [1, 2, 0, 0, 1, 2, 4, 1, 1],
            [0, 0, 2, 0, 1, 3, 5, 0, 1],
            [0, 1, 0, 1, 1, 3, 5, 0, 1],
            [0, 1, 1, 0, 1, 3, 4, 1, 1],
            [0, 2, 0, 0, 1, 3, 3, 2, 1],
        ],
        dtype=int,
    )
    ns_std = np.array(sorted(ns_std.tolist()), dtype=int)
    npt.assert_array_equal(ns, ns_std)

    a = np.array(
        [
            [1, 2, 3, 4, 5, 0, -2, -1, -1],
            [1, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 1, 1, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1],
        ]
    )
    b = np.array([0, 3, 2, 1, 5, 1]) * 3
    n0, vs = solve_diophantines(a, b)
    xs = get_natural_solutions(n0, vs, step=3)
    ns = xs @ vs + n0
    ns = np.array(sorted(ns.tolist()), dtype=int)
    # Scalability test.
    ns_std = (
        np.array(
            [
                [2, 2, 0, 0, 1, 1, 5, 0, 1],
                [1, 1, 1, 0, 1, 2, 5, 0, 1],
                [1, 2, 0, 0, 1, 2, 4, 1, 1],
                [0, 0, 2, 0, 1, 3, 5, 0, 1],
                [0, 1, 0, 1, 1, 3, 5, 0, 1],
                [0, 1, 1, 0, 1, 3, 4, 1, 1],
                [0, 2, 0, 0, 1, 3, 3, 2, 1],
            ],
            dtype=int,
        )
        * 3
    )
    ns_std = np.array(sorted(ns_std.tolist()), dtype=int)
    npt.assert_array_equal(ns, ns_std)

    a = np.array([[0, 0, 0, 0, 0], [1, 1, 1, 1, 1]])
    b = np.array([0, 1])
    n0, vs = solve_diophantines(a, b)
    xs = get_natural_solutions(n0, vs)
    ns = xs @ vs + n0
    ns = np.array(sorted(ns.tolist()), dtype=int)
    # Manually obtained solutions.
    ns_std = np.eye(5, dtype=int)
    ns_std = np.array(sorted(ns_std.tolist()), dtype=int)
    npt.assert_array_equal(ns, ns_std)


def test_flip_size():
    # Test bad flips.
    for _ in range(100):
        u = np.random.randint(low=-100, high=100, size=50)
        if u.sum() != 0:
            with pytest.raises(ValueError):
                _ = flip_size(u)
        else:
            assert flip_size(u) == u[u > 0].sum()


def test_row_matches():
    for _ in range(10):
        a = np.random.randint(low=-100, high=100, size=(100, 100))
        a_clean = np.array(list({tuple(r) for r in a}))
        n = len(a_clean)
        aa1 = a[
            np.random.choice(
                np.arange(2 * n // 3, dtype=int), size=2 * n // 3, replace=False
            )
        ]
        aa2 = a[
            np.random.choice(
                np.arange(n // 3, n, dtype=int), size=n - n // 3, replace=False
            )
        ]
        assert count_row_matches(aa1, aa2) == n // 3


def test_connectivity():
    # Pretty dumb pre-computed test.
    ns = np.array(
        [
            [4, 0, 2, 0, 6],
            [3, 3, 0, 0, 6],
            [3, 2, 1, 1, 5],
            [3, 1, 2, 2, 4],
            [2, 4, 0, 2, 4],
            [3, 0, 3, 3, 3],
            [2, 3, 1, 3, 3],
            [2, 2, 2, 4, 2],
            [1, 5, 0, 4, 2],
            [2, 1, 3, 5, 1],
            [1, 4, 1, 5, 1],
            [2, 0, 4, 6, 0],
            [1, 3, 2, 6, 0],
            [0, 6, 0, 6, 0],
        ],
        dtype=int,
    )
    u1 = np.array([-1, 2, -1, 1, -1], dtype=int)
    assert connectivity(u1, ns) == 8
    assert connectivity(-u1, ns) == 8

    u2 = np.array([0, -1, 1, 1, -1])
    assert connectivity(u2, ns) == 9
    assert connectivity(u2, ns) == 9

    assert connectivity(np.zeros(5, dtype=int), ns) == len(ns)
    assert connectivity([4, -6, 2, -6, 6], ns) == 1


def test_optimal_basis_specific():
    # Random matrices may not be bounded. Do specific tests only.
    # Do 3 pre-computed tests.
    # Test 1: LMTPO, loop also checked. Optimization went well.
    a = np.array([[1, 3, 4, -3, -2], [1, 1, 1, 0, 0], [0, 0, 0, 1, 1]])
    b = np.array([0, 6, 6])
    n0, vs = solve_diophantines(a, b)
    xs = get_natural_solutions(n0, vs)
    vs_opt = get_optimal_basis(n0, vs, xs)
    vs_std = np.array([[0, -1, 1, 1, -1], [-1, 1, 0, 2, -2]])
    assert_table_set_equal(vs_opt, vs_std)

    # Test 2: LNMTOF
    a = np.array(
        [
            [1, 2, 3, 4, -2, -1],
            [1, 1, 1, 1, 0, 0],
            [0, 0, 0, 0, 1, 1],
            [0, 1, -1, 0, 0, 0],
            [0, 0, 1, -1, 0, 0],
        ]
    )
    b = np.array([0, 1, 1, 0, 0])
    n0, vs = solve_diophantines(a, b * 6)
    xs = get_natural_solutions(n0, vs)
    vs_opt = get_optimal_basis(n0, vs, xs)
    vs_std = np.array([[-3, 1, 1, 1, 6, -6]])

    assert_table_set_equal(vs_opt, vs_std)

    # Test 3: Random Alloy
    a = np.array([[0, 0, 0, 0], [1, 1, 0, 0], [0, 0, 1, 1]])
    b = np.array([0, 1, 1])
    n0, vs = solve_diophantines(a, b * 1)
    xs = get_natural_solutions(n0, vs)
    vs_opt = get_optimal_basis(n0, vs, xs)
    vs_std = np.array([[-1, 1, 0, 0], [0, 0, -1, 1]])

    assert_table_set_equal(vs_opt, vs_std)


def test_ergodic_vectors_specific():
    # Random matrices may not be bounded. Do specific tests only.
    # Do 1 pre-computed test.
    # Test 1: LMTOF
    # Pre-computed test
    a = np.array([[1, 3, 4, -2, -1], [1, 1, 1, 0, 0], [0, 0, 0, 1, 1]])
    b = np.array([0, 6, 6])
    n0, vs = solve_diophantines(a, b)
    xs = get_natural_solutions(n0, vs)
    vs_opt = get_optimal_basis(n0, vs, xs)
    len(vs)
    xs_opt = get_natural_solutions(n0, vs_opt)
    vs_tab = get_ergodic_vectors(n0, vs_opt, xs_opt)
    vs_std = np.array([[0, -1, 1, 1, -1], [-1, 1, 0, 2, -2]])
    assert_table_set_equal(vs_tab, vs_std)

    # A non-ergodic table made ergodic.
    n02 = np.array([6, 0, 0, 0, 6], dtype=int)
    vs2 = np.array([[0, -1, 1, 1, -1], [-1, 2, -1, 1, -1]], dtype=int)
    xs2 = np.array([[0, 0], [1, 1], [2, 1], [2, 2], [3, 2], [3, 3], [4, 2]], dtype=int)
    assert np.all(xs2 @ vs2 + n02 >= 0)
    vs_tab2 = get_ergodic_vectors(n02, vs2, xs2)
    vs_std2 = np.array([[0, -1, 1, 1, -1], [-1, 2, -1, 1, -1], [-1, 1, 0, 2, -2]])
    assert_table_set_equal(vs_tab2, vs_std2)


def test_mask():
    for _ in range(10):
        vs = np.random.randint(low=-100, high=100, size=(30, 50))
        table = np.concatenate([(u, -u) for u in vs], axis=0)
        n = np.random.randint(low=0, high=100, size=50)
        aa = np.random.randint(low=-50, high=50, size=50)
        aa @ n
        max_n = np.random.randint(low=100, high=200, size=50)
        mask = flip_weights_mask(vs, n)
        assert len(mask) == len(table)
        assert np.all(np.any(table[~mask, :] + n < 0, axis=-1))
        assert np.all(np.all(table[mask, :] + n >= 0, axis=-1))

        mask = flip_weights_mask(vs, n, max_n=max_n)
        assert np.all(
            np.any(table[~mask, :] + n < 0, axis=-1)
            | np.any(table[~mask, :] + n > max_n, axis=-1)
        )
        assert np.all(
            np.all(table[mask, :] + n >= 0, axis=-1)
            & np.any(table[mask, :] + n <= max_n, axis=-1)
        )

        mask = flip_weights_mask(vs, n, max_n=300)
        assert np.all(np.any(table[~mask, :] + n < 0, axis=-1))
        assert np.all(np.all(table[mask, :] + n >= 0, axis=-1))


def test_choose_sections():
    counts = Counter()
    p = [0.1, 0, 0.3, 0.2, 0.1, 0.3]
    for _ in range(10000):
        counts[choose_section_from_partition(p)] += 1

    for i in range(len(p)):
        # assert abs(counts[i] / 10000 - p[i]) <= 0.05
        # May be this is not required as long as you trust
        # np.random
        if i == 1:
            assert counts[i] <= 1


@pytest.mark.parametrize("determinant", range(6, 13))
def test_yield_hermite_normal_forms(determinant):
    hnfs = [hnf for hnf in yield_hermite_normal_forms(determinant)]
    for hnf in hnfs:
        assert np.linalg.det(hnf) == pytest.approx(determinant)

    # make sure that all the HNFs are unique
    assert len(np.unique(hnfs, axis=0)) == len(hnfs)



"""General tools for structure and occupancy generation."""

from smol.capp.generate.enumerate import enumerate_supercell_matrices
from smol.capp.generate.groundstate import PeriodicGroundStateSolver
from smol.capp.generate.random import generate_random_ordered_occupancy
from smol.capp.generate.special.sqs import StochasticSQSGenerator

__all__ = [
    "enumerate_supercell_matrices",
    "generate_random_ordered_occupancy",
    "StochasticSQSGenerator",
    "PeriodicGroundStateSolver",
]



"""Ground state solver functionalities."""

from monty.dev import requires

try:
    from smol.capp.generate.groundstate.upper_bound.solver import (
        PeriodicGroundStateSolver,
    )
except ImportError:

    @requires(
        False,
        "Ground state solver functionality requires cvxpy to be installed, please install it.",
    )
    class PeriodicGroundStateSolver:
        """Dummy class to fail gracefully when cvxpy is not installed."""


__all__ = ["PeriodicGroundStateSolver"]



"""Implementation of target feature processors.

Feature distance processors are used to compute a distance between the feature values
of a particular configuration and a target feature vector. This functionality is useful
to obtain special structures, such as SQS, SSOS, etc.
"""

__author__ = "Luis Barroso-Luque"

from abc import ABCMeta, abstractmethod
from itertools import chain

import numpy as np

from smol.cofe.space.clusterspace import ClusterSubspace
from smol.moca.processor import ClusterDecompositionProcessor, ClusterExpansionProcessor
from smol.moca.processor.base import Processor


class DistanceProcessor(Processor, metaclass=ABCMeta):
    """Abstract class to compute distance from a fixed feature vector.

    The distance used to measure distance is,

    .. math::

        d = -wL + ||W^T(f - f_T)||_1

    where f is the feature vector, f_T is the target feature vector, and L is the
    diameter for which all values of clusters of smaller diameter are exactly
    equal to the target, and w is a weight for the L term. And W is a diagonal matrix
    of optional weights for each feature.

    As proposed in the following publication,
    https://doi.org/10.1016/j.calphad.2013.06.006
    """

    def __init__(
        self,
        cluster_subspace,
        supercell_matrix,
        target_vector,
        match_weight=1.0,
        match_tol=1e-5,
        target_weights=None,
        **processor_kwargs,
    ):
        """Initialize a DistanceProcessor.

        Args:
            cluster_subspace (ClusterSubspace):
                a cluster subspace
            supercell_matrix (ndarray):
                an array representing the supercell matrix with respect to the
                Cluster Expansion prim structure.
            target_vector (ndarray): optional
                target feature vector (excluding constant)
            match_weight (float): optional
                weight for the in the wL term above. That is how much to weight the
                largest diameter below which all features are matched exactly.
                Set to any number >0 to use this term. Default is 1.0. to ignore it
                set it to zero.
            match_tol (float): optional
                tolerance for matching features. Default is 1e-5.
            target_weights (ndarray): optional
                Weights for the absolute differences each feature when calculating
                the total distance. If None, then all features are weighted equally.
            processor_kwargs (dict): optional
                additional keyword arguments to instantiate the underlying processor
                being inherited from.
        """
        if len(cluster_subspace.external_terms) > 0:
            raise ValueError("The given cluster subspace cannot have external terms.")

        if match_weight < 0:  # maybe we can just allow this, at the users peril?
            raise ValueError("The match weight must be a positive number.")

        if len(target_weights) != len(target_vector) - 1:
            raise ValueError(
                f"The length of target_weights must be equal to the length of"
                f"the target vector minus one {len(target_vector) - 1}. \n"
                f"Got {len(target_weights)} instead."
            )

        self.target_vector = target_vector
        self.match_tol = match_tol

        super().__init__(
            cluster_subspace,
            supercell_matrix,
            coefficients=np.concatenate([[-match_weight], target_weights]),
            **processor_kwargs,
        )

    @abstractmethod
    def compute_feature_vector_distances(self, occupancy, flips):
        """Compute distance of feature vectors separated by given flips from the target.

        Args:
            occupancy (ndarray):
                encoded occupancy array
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a
                single flip where the first element is the index of the site
                in the occupancy array and the second element is the index
                for the new species to place at that site.

        Returns:
            ndarray: 2D distance vector where the first row corresponds to correlations
            before flips
        """
        return

    @abstractmethod
    def exact_match_max_diameter(self, distance_vector):
        """
        Get the largest diameter for which all features are exactly matched.

        Args:
            distance_vector (ndarray):
                feature vector

        Returns:
            float: largest diameter for which all features are exactly matched.
        """
        return

    def compute_feature_vector(self, occupancy):
        """
        Compute the change in the feature vector from a list of flips.

        Args:
            occupancy (ndarray):
                encoded occupancy array

        Returns:
            array: change in feature vector
        """
        feature_vector = (
            super().compute_feature_vector(occupancy) / self.size
        )  # remove scaling
        feature_vector[:] = np.abs(feature_vector - self.target_vector)

        feature_vector[0] = (
            self.exact_match_max_diameter(feature_vector) if self.coefs[0] != 0 else 0.0
        )

        return feature_vector

    def compute_feature_vector_change(self, occupancy, flips):
        """
        Compute the change in the feature vector from a list of flips.

        The feature vector in this case is normalized per supercell. In
        other works it is extensive corresponding to the size of the supercell.

        Args:
            occupancy (ndarray):
                encoded occupancy string
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a
                single flip where the first element is the index of the site
                in the occupancy string and the second element is the index
                for the new species to place at that site.

        Returns:
            array: change in feature vector
        """
        distance_vectors = self.compute_feature_vector_distances(occupancy, flips)

        if self.coefs[0] != 0:
            distance_vectors[0, 0] = self.exact_match_max_diameter(distance_vectors[0])
            distance_vectors[1, 0] = self.exact_match_max_diameter(distance_vectors[1])

        return distance_vectors[1] - distance_vectors[0]

    def as_dict(self):
        """Return a dictionary representation of the processor."""
        d = super().as_dict()
        d["target_vector"] = self.target_vector.tolist()
        d["match_weight"] = -self.coefs[0]
        return d

    @classmethod
    def from_dict(cls, d):
        """Create a processor from a dictionary representation."""
        cluster_subspace = ClusterSubspace.from_dict(d.pop("cluster_subspace"))
        supercell_matrix = np.array(d.pop("supercell_matrix"))
        target_vector = np.array(d.pop("target_vector"))
        match_weight = d.pop("match_weight")

        # remove @ and coefficient values:
        d = {k: v for k, v in d.items() if "@" not in k and k != "coefficients"}
        return cls(
            cluster_subspace,
            supercell_matrix=supercell_matrix,
            target_vector=target_vector,
            match_weight=match_weight,
            **d,
        )


class CorrelationDistanceProcessor(DistanceProcessor, ClusterExpansionProcessor):
    """CorrelationDistanceProcessor to compute distance from a fixed correlation vector.

    The distance used to measure distance is,

    .. math::

        d = -wL + ||W^T(f - f_T)||_1

    where f is the correlation vector, f_T is the target correlation vector, and L is the
    diameter for which all correlation values of clusters of smaller diameter are exactly
    equal to the target, and w is a weight for the L term. And W is a diagonal matrix
    of optional weights for each correlation.

    As proposed in the following publication,
    https://doi.org/10.1016/j.calphad.2013.06.006
    """

    def __init__(
        self,
        cluster_subspace,
        supercell_matrix,
        target_vector=None,
        match_weight=1.0,
        target_weights=None,
        match_tol=1e-5,
    ):
        """Initialize a CorrelationDistanceProcessor.

        Args:
            cluster_subspace (ClusterSubspace):
                a cluster subspace
            supercell_matrix (ndarray):
                an array representing the supercell matrix with respect to the
                Cluster Expansion prim structure.
            target_vector (ndarray): optional
                target correlation vector, if None given as vector of zeros is used.
            match_weight (float): optional
                weight for the in the wL term above. That is how much to weight the
                largest diameter below which all correlations are matched exactly.
                Set to any number >0 to use this term. Default is 1.0. to ignore it
                set it to zero.
            match_tol (float): optional
                tolerance for matching features. Default is 1e-8.
            target_weights (ndarray): optional
                Weights for the absolute differences each correlation when calculating
                the total distance. If None, then all correlations are weighted equally.
        """
        if target_vector is None:
            target_vector = np.zeros(len(cluster_subspace))
        if target_weights is None:
            target_weights = np.ones(len(cluster_subspace) - 1)

        super().__init__(
            cluster_subspace,
            supercell_matrix,
            target_vector=target_vector,
            match_weight=match_weight,
            target_weights=target_weights,
            match_tol=match_tol,
        )

    def compute_feature_vector_distances(self, occupancy, flips):
        """Compute the distance of correlation vectors separated by given flips.

        The distance is referenced to the target correlation vector, such that the
        result is a 2D array where each row is the distance of one correlation vector
        to the target.

        Args:
            occupancy (ndarray):
                encoded occupancy array
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a single flip
                where the first element is the index of the site in the occupancy array
                and the second element is the index for the new species to place at that
                site.

        Returns:
            ndarray: 2D distance vector where the first row corresponds to correlations
            before flips
        """
        occu_f = occupancy.copy()
        for f in flips:
            occu_f[f[0]] = f[1]

        corr_distances = self._evaluator.corr_distances_from_occupancies(
            occu_f, occupancy, self.target_vector, self._indices.container
        )
        return corr_distances

    def exact_match_max_diameter(self, distance_vector):
        """
        Get the largest diameter for which all features are exactly matched.

        Args:
            distance_vector (ndarray):
                feature vector

        Returns:
            float: largest diameter
        """
        max_matched_diameter = 0.0

        for diameter, orbits in self.cluster_subspace.orbits_by_diameter.items():
            indices = list(
                chain.from_iterable(
                    range(orb.bit_id, orb.bit_id + len(orb)) for orb in orbits
                )
            )
            if np.all(distance_vector[indices] <= self.match_tol):
                max_matched_diameter = diameter
            else:
                break

        return max_matched_diameter


class ClusterInteractionDistanceProcessor(
    DistanceProcessor, ClusterDecompositionProcessor
):
    """Compute distances from a fixed cluster interaction vector.

    The distance used to measure distance is,

    .. math::

        d = -wL + ||W^T(f - f_T)||_1

    where f is the cluster interaction vector, f_T is the target vector, and L is the
    diameter for which all correlation values of clusters of smaller diameter are
    exactly equal to the target, and w is a weight for the L term. And W is a diagonal
    matrix of optional weights for each correlation.

    Following the proposal in the following publication but applied to cluster
    interaction vectors,
    https://doi.org/10.1016/j.calphad.2013.06.006
    """

    def __init__(
        self,
        cluster_subspace,
        supercell_matrix,
        interaction_tensors=None,
        target_vector=None,
        match_weight=1.0,
        target_weights=None,
        match_tol=1e-5,
    ):
        """Initialize a CorrelationDistanceProcessor.

        Args:
            cluster_subspace (ClusterSubspace):
                a cluster subspace
            supercell_matrix (ndarray):
                an array representing the supercell matrix with respect to the
                Cluster Expansion prim structure.
            interaction_tensors (sequence of ndarray): optional
                Sequence of ndarray where each array corresponds to the
                cluster interaction tensor. These should be in the same order as their
                corresponding orbits in the given cluster_subspace
            target_vector (ndarray): optional
                target cluster interaction vector, if None given as vector of zeros is
                used.
            match_weight (float): optional
                weight for the in the wL term above. That is how much to weight the
                largest diameter below which all correlations are matched exactly.
                Set to any number >0 to use this term. Default is 1.0. to ignore it
                set it to zero.
            match_tol (float): optional
                tolerance for matching features. Default is 1e-8.
            target_weights (ndarray): optional
                Weights for the absolute differences each correlation when calculating
                the total distance. If None, then all correlations are weighted equally.
        """
        if target_vector is None:
            target_vector = np.zeros(cluster_subspace.num_orbits)

        if target_weights is None:
            target_weights = np.ones(cluster_subspace.num_orbits - 1)

        # if none given, compute cluster interactions with all coefficients set to 1
        if interaction_tensors is None:
            interaction_tensors = (0.0,) + tuple(
                sum(
                    m * tensor
                    for i, (m, tensor) in enumerate(
                        zip(orbit.bit_combo_multiplicities, orbit.correlation_tensors)
                    )
                )
                for orbit in cluster_subspace.orbits
            )
        interaction_tensors = tuple(
            np.ascontiguousarray(interaction) for interaction in interaction_tensors
        )

        super().__init__(
            cluster_subspace,
            supercell_matrix,
            interaction_tensors=interaction_tensors,
            target_vector=target_vector,
            match_weight=match_weight,
            target_weights=target_weights,
            match_tol=match_tol,
        )

    def compute_feature_vector_distances(self, occupancy, flips):
        """Compute the distance of cluster interaction vectors separated by given flips.

        Args:
            occupancy (ndarray):
                encoded occupancy array
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a single flip
                where the first element is the index of the site in the occupancy array
                and the second element is the index for the new species to place at that
                site.

        Returns:
            ndarray: 2D distance vector where the first row corresponds to correlations
            before flips
        """
        occu_f = occupancy.copy()
        for f in flips:
            occu_f[f[0]] = f[1]
        interaction_distances = self._evaluator.interaction_distances_from_occupancies(
            occu_f, occupancy, self.target_vector, self._indices.container
        )
        return interaction_distances

    def exact_match_max_diameter(self, distance_vector):
        """
        Get the largest diameter for which all features are exactly matched.

        Args:
            distance_vector (ndarray):
                feature vector

        Returns:
            float: largest diameter
        """
        max_matched_diameter = 0.0

        for diameter, orbits in self.cluster_subspace.orbits_by_diameter.items():
            indices = [orb.id for orb in orbits]
            if np.all(distance_vector[indices] <= self.match_tol):
                max_matched_diameter = diameter
            else:
                break

        return max_matched_diameter



import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from smol.capp.generate import enumerate_supercell_matrices


@pytest.mark.parametrize("size", range(2, 9, 2))
def test_enumerate_supercell_matrices(structure, size):
    symops = SpacegroupAnalyzer(structure).get_symmetry_operations()
    scms = enumerate_supercell_matrices(size, symops)

    # assert determinants are correct
    for scm in scms:
        assert np.linalg.det(scm) == pytest.approx(size)

    # make sure that all the scms are unique
    assert len(np.unique(scms, axis=0)) == len(scms)

    # check that no two matrices are related by symmetry
    for scm in scms:
        for symop in symops:
            rot = np.linalg.inv(scm.T) @ symop.rotation_matrix
            equiv = [(abs(rot @ m.T - np.round(rot @ m.T)) < 1e-5).all() for m in scms]
            assert sum(equiv) <= 1  # at most one matrix is equivalent (ie itself)
            if sum(equiv) == 1:  # make sure the equivalent is actually the same
                equiv_scm = scms[equiv.index(True)]
                npt.assert_allclose(scm, equiv_scm)



"""
A few testing utilities that may be useful to just import and run.
Some of these are borrowed from pymatgen test scripts.
"""

import json
import pickle

import numpy as np
import numpy.testing as npt
from monty.json import MontyDecoder, MSONable
from pymatgen.core import Composition
from pymatgen.entries.computed_entries import ComputedStructureEntry


def assert_table_set_equal(a1, a2):
    a1 = np.array(a1)
    a2 = np.array(a2)
    assert len(a1.shape) == 2
    assert a1.shape == a2.shape
    a1_set = np.concatenate((a1, -a1), axis=0)
    a2_set = np.concatenate((a2, -a2), axis=0)
    a1_set = np.array(sorted(tuple(r) for r in a1_set))
    a2_set = np.array(sorted(tuple(r) for r in a2_set))
    npt.assert_array_equal(a1_set, a2_set)


def assert_msonable(obj, skip_keys=None, test_if_subclass=True):
    """
    Tests if obj is MSONable and tries to verify whether the contract is
    fulfilled.
    By default, the method tests whether obj is an instance of MSONable.
    This check can be deactivated by setting test_if_subclass to False.
    """
    if test_if_subclass:
        assert isinstance(obj, MSONable)

    skip_keys = [] if skip_keys is None else skip_keys
    d1 = obj.as_dict()
    d2 = obj.__class__.from_dict(obj.as_dict()).as_dict()
    for key in d1.keys():
        if key in skip_keys:
            continue
        # if d1[key] != d2[key]:
        #     print("key:", key)
        #     print("d1:", d1[key])
        #     print("d2:", d2[key])

        assert d1[key] == d2[key]

    try:
        _ = json.loads(obj.to_json(), cls=MontyDecoder)
    except Exception as e:
        raise AssertionError(e)


def assert_pickles(obj):
    """Test if obj is picklable."""
    try:
        p = pickle.dumps(obj)
        obj_copy = pickle.loads(p)
    except Exception as e:
        raise AssertionError(e)

    assert isinstance(obj_copy, obj.__class__)

    if isinstance(obj, MSONable):
        d1 = obj.as_dict()
        d2 = obj_copy.as_dict()
        for key in d1.keys():
            assert d1[key] == d2[key]
    else:
        # fallback for objects that are not MSONable
        # not a complete test, since we are only checking that attribute names match
        d1 = obj.__dict__
        d2 = obj_copy.__dict__
        assert d1.keys() == d2.keys()


def gen_fake_training_data(prim_structure, n=10, rng=None):
    """Generate a fake structure, energy training set."""
    rng = np.random.default_rng(rng)
    training_data = []
    for energy in rng.random(n):
        struct = gen_random_ordered_structure(
            prim_structure, size=rng.integers(2, 6), rng=rng
        )
        energy *= -len(struct)
        training_data.append(ComputedStructureEntry(struct, energy))
    return training_data


def gen_random_ordered_structure(prim, size=3, rng=None):
    """Generate an random ordered structure from a disordered prim.

    Args:
        prim (pymatgen.Structure):
            disordered primitive structure:
        size (optional):
            size argument to structure.make_supercell
        rng (optional): {None, int, array_like[ints], SeedSequence, BitGenerator, Generator},
            A RNG, seed or otherwise to initialize defauly_rng

    Returns:
        ordered structure
    """
    rng = np.random.default_rng(rng)
    structure = prim.copy()
    structure.make_supercell(size)
    for site in structure:
        site.species = Composition({rng.choice(list(site.species.keys())): 1})
    return structure


def compute_cluster_interactions(expansion, structure, normalized=True, scmatrix=None):
    # compute cluster interactions from definitionas in pure python
    corrs = expansion.cluster_subspace.corr_from_structure(
        structure, normalized=normalized, scmatrix=scmatrix
    )
    vals = expansion.eci * corrs
    interactions = np.array(
        [
            np.sum(
                vals[expansion.eci_orbit_ids == i]
                * expansion.cluster_subspace.function_ordering_multiplicities[
                    expansion.eci_orbit_ids == i
                ]
            )
            for i in range(len(expansion.cluster_subspace.orbits) + 1)
        ]
    )
    return interactions



from dataclasses import dataclass, field
from typing import Dict, List, Union

from smol.utils.setmany import SetMany


@dataclass
class A:
    index: int
    num_threads: int


@dataclass
class Container:
    many: Union[List[A], Dict[int, A]]
    num_threads: int = field(default=SetMany("num_threads", "many"))


def test_set_many():
    # test with list
    many = [A(i, 1) for i in range(10)]
    test = Container(many, 1)

    assert test.num_threads == 1
    for a in test.many:
        assert a.num_threads == 1

    test.num_threads = 2

    assert test.num_threads == 2
    for a in test.many:
        assert a.num_threads == 2

    # test with dict
    many = {i: A(i, 1) for i in range(10)}
    test = Container(many, 1)

    assert test.num_threads == 1
    for a in test.many.values():
        assert a.num_threads == 1

    test.num_threads = 2

    assert test.num_threads == 2
    for a in test.many.values():
        assert a.num_threads == 2



"""Special Quasi-Random Structures (SQS)."""

__author__ = "Luis Barroso-Luque"


import warnings
from abc import ABC, abstractmethod
from collections import deque, namedtuple
from copy import deepcopy

import numpy as np
from pymatgen.analysis.structure_matcher import StructureMatcher

from smol.capp.generate.enumerate import enumerate_supercell_matrices
from smol.capp.generate.random import generate_random_ordered_occupancy
from smol.cofe import ClusterSubspace
from smol.moca import Ensemble, SampleContainer, Sampler
from smol.moca.kernel import MulticellMetropolis, mckernel_factory
from smol.moca.processor.distance import (
    ClusterInteractionDistanceProcessor,
    CorrelationDistanceProcessor,
    DistanceProcessor,
)
from smol.moca.trace import Trace
from smol.utils.class_utils import class_name_from_str, derived_class_factory
from smol.utils.progressbar import progress_bar

SQS = namedtuple("SQS", ["structure", "score", "feature_distance", "supercell_matrix"])


class SQSGenerator(ABC):
    """Abstract SQSGenerator class."""

    def __init__(
        self,
        cluster_subspace,
        supercell_size,
        feature_type="correlation",
        target_vector=None,
        target_weights=None,
        match_weight=1.0,
        match_tol=1e-5,
        supercell_matrices=None,
        processors=None,
        **kwargs,
    ):
        """Initialize SQSGenerator.

        Args:
            cluster_subspace (ClusterSubspace):
                cluster subspace used to determine feature distance
            supercell_size (int):
                size of the supercell in multiples of the primitive cell
            feature_type (str): optional
                type of features to be used to determine SQS.
                options are "correlation" and "cluster-interaction".
            target_vector (ndarray): optional
                target feature vector to use for distance calculations
            target_weights (ndarray): optional
                weights for each target feature
            match_weight (float): optional
                weight for the in the wL term above. That is how much to weight the
                largest diameter below which all features are matched exactly.
                Set to any number >0 to use this term. Default is 1.0. to ignore it
                set it to zero.
            match_tol (float): optional
                tolerance for matching features. Default is 1e-5.
            supercell_matrices (list of ndarray): optional
                list of supercell matrices to use. If None, all symmetrically distinct
                supercell matrices are generated.
            processors (list of DistanceProcessor): optional
                list of processors to use. If this is passed supercell_matrices
                cannot be passed.
        """
        self.cluster_subspace = cluster_subspace
        self.supercell_size = supercell_size
        self._sqs_deque = None

        if processors is not None:
            if processors[0].cluster_subspace != cluster_subspace:
                raise ValueError(
                    "Processors must have the same cluster subspace as the generator"
                )
            if not all(processor.size == supercell_size for processor in processors):
                raise ValueError(
                    "processors must have the same supercell size given to the generator"
                )

        if feature_type == "correlation":
            num_features = len(cluster_subspace)
        elif feature_type == "cluster-interaction":
            num_features = cluster_subspace.num_orbits
        else:
            raise ValueError(
                f"feature_type {feature_type} not supported.\n "
                f"Valid options are: 'correlation', 'cluster-interaction'"
            )

        if target_weights is None:
            target_weights = np.ones(num_features - 1)  # remove constant
        else:
            if len(target_weights) != num_features - 1:
                raise ValueError(f"target_weights must be of length {num_features - 1}")

        if target_vector is None:
            target_vector = np.zeros(num_features)
        else:
            if len(target_vector) != num_features:
                raise ValueError(
                    f"target feature vector must be of length {num_features}"
                )

        if supercell_matrices is not None:
            if processors is not None:
                raise ValueError(
                    "processors and supercell_matrices cannot both be specified"
                )
            for scm in supercell_matrices:
                if scm.shape != (3, 3):
                    raise ValueError("supercell matrices must be 3x3")
                if not np.isclose(np.linalg.det(scm), supercell_size):
                    raise ValueError(
                        "supercell matrices must have determinant equal to supercell_size"
                    )
        elif processors is None:
            supercell_matrices = enumerate_supercell_matrices(
                supercell_size, cluster_subspace.symops
            )
            # reverse since most skewed cells are enumerated first
            supercell_matrices.reverse()

        if processors is None:
            self._processors_by_scm = {
                tuple(sorted(tuple(s.tolist()) for s in scm)): derived_class_factory(
                    class_name_from_str(feature_type + "DistanceProcessor"),
                    DistanceProcessor,
                    cluster_subspace,
                    scm,
                    target_vector=target_vector,
                    target_weights=target_weights,
                    match_weight=match_weight,
                    match_tol=match_tol,
                )
                for scm in supercell_matrices
            }
        else:
            # TODO should we also check the type of processor?
            self._processors_by_scm = {
                tuple(
                    sorted(tuple(s.tolist()) for s in processor.supercell_matrix)
                ): processor
                for processor in processors
            }

        self._processors = list(self._processors_by_scm.values())

    @classmethod
    def from_structure(
        cls,
        structure,
        cutoffs,
        supercell_size,
        basis="indicator",
        orthonormal=True,
        use_concentration=True,
        feature_type="correlation",
        target_vector=None,
        target_weights=None,
        match_weight=1.0,
        match_tol=1e-5,
        supercell_matrices=None,
        **kwargs,
    ):
        """Initialize an SQSGenerater from a disordered structure.

        Args:
            structure (Structure):
                a disordered pymatgen structure
            cutoffs (dict of int: float):
                cluster diameter cutoffs
            supercell_size (int):
                size of the supercell in multiples of the primitive cell
            basis (str):
                a string specifying the site basis functions
            orthonormal (bool):
                whether to enforce an orthonormal basis. From the current
                available bases only the indicator basis is not orthogonal
                out-of-the-box
            use_concentration (bool):
                if True, the concentrations in the prim structure sites will be
                used to orthormalize site bases. This gives a cluster
                subspace centered about the prim composition.
            feature_type (str): optional
                type of features to be used to determine SQS.
                options are: "correlation"
            target_vector (ndarray): optional
                target feature vector to use for distance calculations
            target_weights (ndarray): optional
                weights for each target feature
            match_weight (float): optional
                weight for the in the wL term above. That is how much to weight the
                largest diameter below which all features are matched exactly.
                Set to any number >0 to use this term. Default is 1.0. to ignore it
                set it to zero.
            match_tol (float): optional
                tolerance for exact matching features. Default is 1e-5.
            supercell_matrices (list of ndarray): optional
                list of supercell matrices to use. If None, all symmetrically distinct
                supercell matrices are generated.
            **kwargs:
                additional keyword arguments specific to the SQSGenerator

        Returns:
            SQSGenerator
        """
        subspace = ClusterSubspace.from_cutoffs(
            structure,
            cutoffs,
            basis=basis,
            orthonormal=orthonormal,
            use_concentration=use_concentration,
        )
        return cls(
            subspace,
            supercell_size,
            feature_type=feature_type,
            target_vector=target_vector,
            target_weights=target_weights,
            match_weight=match_weight,
            match_tol=match_tol,
            supercell_matrices=supercell_matrices,
            **kwargs,
        )

    @classmethod
    def from_processors(
        cls,
        processors,
        target_vector=None,
        target_weights=None,
        match_weight=1.0,
        match_tol=1e-5,
        **kwargs,
    ):
        """Initialize an SQSGenerator from a list of processors.

        Args:
            processors: list of processors
            target_vector (ndarray): optional
                target feature vector to use for distance calculations
            target_weights (ndarray): optional
                weights for each target feature
            match_weight (float): optional
                weight for the in the wL term above. That is how much to weight the
                largest diameter below which all features are matched exactly.
                Set to any number >0 to use this term. Default is 1.0. to ignore it
                set it to zero.
            match_tol (float): optional
                tolerance for exact matching features. Default is 1e-5.
            **kwargs:
                additional keyword arguments specific to the SQSGenerator
        Returns:
            SQSGenerator
        """
        if isinstance(processors[0], CorrelationDistanceProcessor):
            feature_type = "correlation"
        elif isinstance(processors[0], ClusterInteractionDistanceProcessor):
            feature_type = "cluster-interaction"
        else:
            raise ValueError(f"Invalid processor type {type(processors[0])}")

        return cls(
            processors[0].cluster_subspace,
            processors[0].size,
            feature_type=feature_type,
            target_vector=target_vector,
            target_weights=target_weights,
            match_weight=match_weight,
            match_tol=match_tol,
            processors=processors,
            **kwargs,
        )

    @property
    def num_structures(self):
        """Get number of structures generated."""
        return 0 if self._sqs_deque is None else len(self._sqs_deque)

    @property
    def processors(self):
        """Get list of processors."""
        return self._processors

    @abstractmethod
    def generate(self, *args, **kwargs):
        """Generate a SQS structures."""
        return

    def compute_score(self, structure, supercell_matrix=None):
        """Compute the SQS score for a structure.

        Args:
            structure (Structure):
                ordered structure to compute SQS score for
            supercell_matrix (ndarray): optional
                supercell matrix relating given ordered structure to the disordered
                structure.

        Returns:
            float : SQS score
        """
        processor = self._get_structure_processor(structure, supercell_matrix)
        occu = processor.occupancy_from_structure(structure)
        return processor.compute_property(occu)

    def compute_feature_distance(self, structure, supercell_matrix=None):
        """Compute feature distance to target vector for a given structure.

        Args:
            structure (Structure):
                ordered structure to compute SQS score for
            supercell_matrix (ndarray): optional
                supercell matrix relating given ordered structure to the disordered
                structure.

        Returns:
            float : feature distance
        """
        processor = self._get_structure_processor(structure, supercell_matrix)
        occu = processor.occupancy_from_structure(structure)
        return processor.compute_feature_vector(occu)

    def _get_structure_processor(self, structure, supercell_matrix):
        """Check if an ordered structure and supercell matrix are valid."""
        if supercell_matrix is None:
            supercell_matrix = self.cluster_subspace.scmatrix_from_structure(structure)

        if not np.isclose(np.linalg.det(supercell_matrix), self.supercell_size):
            raise ValueError(
                "Invalid supercell matrix, must have determinant equal to supercell_size"
            )
        scm = tuple(sorted(tuple(s.tolist()) for s in supercell_matrix))
        return self._processors_by_scm[scm]

    def get_best_sqs(
        self, num_structures=1, remove_duplicates=True, reduction_algorithm="niggli"
    ):
        """Get the best SQS structures.

        generate method must be called, otherwise no SQS will be returned.

        Args:
            num_structures (int): optional
                number of structures to return
            remove_duplicates (bool): optional
                whether to remove duplicate structures
            reduction_algorithm (callable): optional
                function to reduce the number of structures returned
                should take a list of SQS structures and return a list of SQS structures
                options are "niggli" and "LLL".
                See documentation for Structure.get_reduced_structure in pymatgen
                https://pymatgen.org/pymatgen.core.structure.html

        Returns:
            list of SQS : list ranked by SQS score
        """
        if num_structures > self.num_structures:
            warnings.warn(
                f"num_structures is greater than the total number of "
                f"structures generated {self.num_structures}. \n Will return at most "
                f"{self.num_structures} structures."
            )

        best_traces = sorted(self._sqs_deque, key=lambda x: x.enthalpy)

        # package traces into SQS objects
        best_sqs = []
        for trace in best_traces:
            i = np.argmin(trace.enthalpy)  # in case anyone runs multiple walkers
            processor = self._processors[trace.kernel_index[i]]
            structure = processor.structure_from_occupancy(trace.occupancy[i])
            sqs = SQS(
                structure=structure,
                score=trace.enthalpy[i][0],
                feature_distance=trace.features[i],
                supercell_matrix=processor.supercell_matrix,
            )
            best_sqs.append(sqs)

            if num_structures == 1:  # just do one
                break

        if remove_duplicates:
            matcher = StructureMatcher()
            grouped_structures = matcher.group_structures(
                [sqs.structure for sqs in best_sqs]
            )
            unique_structures = [group[0] for group in grouped_structures]
            best_sqs = list(
                filter(lambda x: x.structure in unique_structures, best_sqs)
            )

        if len(best_sqs) < num_structures:
            warnings.warn(
                f"Fewer SQS structures than requested were obtained after removing "
                f"duplicates. Only {len(best_sqs)} distinct SQS will be returned."
            )
        else:
            best_sqs = best_sqs[:num_structures]

        if reduction_algorithm is not None:
            for i, sqs in enumerate(best_sqs):
                structure = sqs.structure.get_reduced_structure(
                    reduction_algo=reduction_algorithm
                )
                reduced_sqs = SQS(
                    structure=structure,
                    score=sqs.score,
                    feature_distance=sqs.feature_distance,
                    supercell_matrix=sqs.supercell_matrix,
                )
                best_sqs[i] = reduced_sqs

        return best_sqs


class StochasticSQSGenerator(SQSGenerator):
    """StochasticSQSGenertor class.

    Generate special quasirandom structures using simulated annealing, based on the
    following work

    https://doi.org/10.1016/j.calphad.2013.06.006
    """

    def __init__(
        self,
        cluster_subspace,
        supercell_size,
        feature_type="correlation",
        target_vector=None,
        target_weights=None,
        match_weight=1.0,
        match_tol=1e-5,
        supercell_matrices=None,
        processors=None,
        kernel_kwargs=None,
        **kwargs,
    ):
        """Initialize StochasticSQSGenerator.

        Args:
            cluster_subspace (ClusterSubspace):
                cluster subspace used to determine feature_distance
            supercell_size (int):
                size of the supercell in multiples of the primitive cell
            feature_type (str): optional
                type of feature_distance to be used to determine SQS.
                options are: "correlation"
            target_vector (ndarray): optional
                target feature vector to use for distance calculations
            target_weights (ndarray): optional
                weights for each target feature
            match_weight (float): optional
                weight for the in the wL term above. That is how much to weight the
                largest diameter below which all features are matched exactly.
                Set to any number >0 to use this term. Default is 1.0. to ignore it
                set it to zero.
            match_tol (float): optional
                tolerance for matching features. Default is 1e-5.
            supercell_matrices (list of ndarray): optional
                list of supercell matrices to use. If None, all symmetrically distinct
                supercell matrices are generated.
            processors (list of Processor): optional
                list of processors to use. If None, all processors are generated.
            kernel_kwargs (dict): optional
                keyword arguments for the transition kernel used in each supercell.
                For example see the documentation for :class:`Metropolis` kernel.
            **kwargs:
                keyword arguments used to initialize the :class:`MulticellMetropolis`
                kernel to be used for the simulated annealing.
        """
        super().__init__(
            cluster_subspace,
            supercell_size,
            feature_type,
            target_vector,
            target_weights,
            match_weight,
            match_tol,
            supercell_matrices,
            processors,
        )
        step_type = kwargs.pop("step_type", "swap")
        temperature = kwargs.pop("temperature", 5.0)

        if step_type != "swap":
            warnings.warn(
                f"Step type {step_type} was provided.\n Swap steps are recommended, "
                f"make sure this is what you want.",
                UserWarning,
            )

        kernel_kwargs = kernel_kwargs or {}
        kernels = [
            mckernel_factory(
                "metropolis",
                Ensemble(processor),
                step_type,
                temperature=temperature,
                **kernel_kwargs,
            )
            for processor in self._processors
        ]
        for kernel in kernels:
            kernel.kB = 1.0  # set kB to 1.0 units
        self._kernel = MulticellMetropolis(kernels, temperature=temperature, **kwargs)
        self._kernel.kB = 1.0  # set kB to 1.0 units

        # get a trial trace to initialize sample container trace
        # Enforce int32.
        _trace = self._kernel.compute_initial_trace(
            np.zeros(kernels[0].ensemble.num_sites, dtype=np.int32)
        )
        sample_trace = Trace(
            **{
                name: np.empty((0, 1, *value.shape), dtype=value.dtype)
                for name, value in _trace.items()
            }
        )

        container = SampleContainer(kernels[0].ensemble, sample_trace)
        container.metadata.type = "SQS-SampleContainer"
        self._sampler = Sampler([self._kernel], container)

    @property
    def sampler(self):
        """Sampler: sampler used to generate SQS."""
        return self._sampler

    def generate(
        self,
        mcmc_steps,
        temperatures=None,
        initial_occupancies=None,
        clear_previous=True,
        max_save_num=None,
        progress=False,
    ):
        """Generate a SQS structures.

        Args:
            mcmc_steps (int):
                number of mcmc steps per temperature to perform simulated annealing
            temperatures (list of float): optional
                temperatures to use for the mcmc steps, temperatures are unit-less
                the recommended range should be only single digits.
            initial_occupancies (ndarray): optional
                initial occupancies to use in the simulated anneal search, an occupancy
                must be given for each of the supercell shapes, and these occupancies
                must have the correct compositions.
            clear_previous (bool): optional
                whether to clear previous samples.
            max_save_num (int): optional
                max number of best structures to save, if None a maximum 1% of
                structures are saved. A structure is only saved if the SQS score is
                better than the previous best.
            progress (bool):
                if true will show a progress bar.
        """
        if initial_occupancies is None:
            if self._sampler.samples.num_samples == 0:
                initial_occupancies = self._get_initial_occupancies()
            else:
                initial_occupancies = self._sampler.samples.get_occupancies(flat=False)[
                    -1
                ]
        elif initial_occupancies is not None:
            shape = (len(self._processors), self._kernel.ensemble.num_sites)
            if initial_occupancies.shape != shape:
                raise ValueError(
                    f"initial occupancies must be of shape {shape}, "
                    f"got {initial_occupancies.shape}"
                )
            initial_occupancies = initial_occupancies.copy()

        max_save_num = max_save_num or max(int(0.01 * mcmc_steps), 1)
        if clear_previous:
            self._sampler.clear_samples()
            self._sqs_deque = deque(maxlen=max_save_num)
        else:
            self._sqs_deque = deque(
                self._sqs_deque, maxlen=len(self._sqs_deque) + max_save_num
            )

        if temperatures is None:
            temperatures = np.linspace(5.0, 0.01, 20)

        self._kernel.temperature = temperatures[0]
        for kernel in self._kernel.mckernels:
            kernel.temperature = temperatures[0]

        best_score = np.inf
        for trace in self._sample_sqs(
            mcmc_steps, initial_occupancies, progress=progress
        ):
            if np.any(trace.enthalpy < best_score):
                best_score = trace.enthalpy.min()
                self._sqs_deque.append(deepcopy(trace))

        for temperature in temperatures[1:]:
            self._kernel.temperature = temperature
            for kernel in self._kernel.mckernels:
                kernel.temperature = temperature
            initial_occupancies = trace.occupancy
            for trace in self._sample_sqs(
                mcmc_steps, initial_occupancies, progress=progress
            ):
                if np.any(trace.enthalpy < best_score):
                    best_score = trace.enthalpy.min()
                    self._sqs_deque.append(deepcopy(trace))

        self._sampler.samples.allocate(max_save_num)
        for trace in self._sqs_deque:
            self._sampler.samples.save_sampled_trace(trace, 1)
        self._sampler.samples.vacuum()  # vacuum since not all space may have been used

    def _sample_sqs(self, nsteps, initial_occupancies, progress=False):
        """Sample SQS structures.

        Very similar to sampler.sample but without thinning and only yielding better
        SQS structures.

        Args:
            nsteps (int):
                number of steps to sample
            progress (bool):
                whether to show a progress bar

        Yields:
            tuple: accepted, occupancies, features change, enthalpies change
        """
        occupancies, trace = self._sampler.setup_sample(initial_occupancies)
        chains, nsites = self._sampler.samples.shape
        desc = f"Generating SQS using {chains} chain(s) from cells with {nsites} sites"
        with progress_bar(progress, total=nsteps, description=desc) as p_bar:
            for _ in range(nsteps):
                # the occupancies are modified in place at each step
                for i, strace in enumerate(
                    map(
                        lambda kernel, occu: kernel.single_step(occu),
                        self._sampler.mckernels,
                        occupancies,
                    )
                ):
                    for name, value in strace.items():
                        val = getattr(trace, name)
                        val[i] = value
                        # this will mess up recording values for > 1 walkers
                        # setattr(trace, name, value)
                    if strace.accepted:
                        for name, delta_val in strace.delta_trace.items():
                            val = getattr(trace, name)
                            val[i] += delta_val
                p_bar.update()
                yield trace

    def _get_initial_occupancies(self):
        """Get initial occupancies for the simulated annealing.

        Gets initial occupancies for each supercell shape at the composition of the
        disordered structure in the given cluster subspace.

        Returns:
            ndarray of occupancies
        """
        # all processors should have the same compositions so just use one processor
        compositions = [sl.composition for sl in self._processors[0].get_sublattices()]
        occupancies = np.vstack(
            [
                generate_random_ordered_occupancy(proc, composition=compositions)
                for proc in self._processors
            ]
        )
        return occupancies



"""Utility manager that converts input of composition constraints to equations."""

import itertools
import re
from numbers import Number

import numpy as np

from smol.cofe.space.domain import get_species
from smol.moca.occu_utils import get_dim_ids_by_sublattice

__author__ = "Fengyu Xie"


def _extract_symbols_before_number(all_literals, number_literal_id):
    """Extract operator symbols and count the number of symbols."""
    num_symbols = 0
    symbol = 1
    if number_literal_id > 0:
        for jj in range(1, number_literal_id + 1):
            if all_literals[number_literal_id - jj] == "+":
                num_symbols += 1
            elif all_literals[number_literal_id - jj] == "-":
                symbol = -1 * symbol
                num_symbols += 1
            else:
                break

    return symbol, num_symbols


def _extract_number_before_species(all_literals, species_literal_id):
    """Extract species string and the amount of species."""
    num_numbers = 0
    if species_literal_id > 0:
        for jj in range(1, species_literal_id + 1):
            if isinstance(all_literals[species_literal_id - jj], Number):
                num_numbers += 1
            else:
                break
    if num_numbers == 0:
        number_literal = 1
    elif num_numbers == 1:
        number_literal = all_literals[species_literal_id - 1]
    else:
        raise ValueError(
            f"Species {all_literals[species_literal_id]} preceded by"
            f" {num_numbers} > 1"
            f" literals, not allowed!"
        )

    return number_literal, num_numbers


def handle_side_string(side: str):
    """Handle the strings separated by relation operators.

    Args:
        side(str):
           A side of equation in string.
    Returns:
        list of tuple of (Number, Species, int or None), Number:
            A list of tuples packing coefficients, species, and sub-lattice indices (if any)
            on this side of equation, and the intercept term.
    """
    # split around spaces. each literal can be either "+", "-", "\d" or species string.
    literals = side.split()
    processed_literals = []

    numerical_reg = re.compile(r"^[+-]?\d+\.?\d*$")
    sublattice_reg = re.compile(r"^([A-Za-z]+.*)\((\d+)\)$")
    for i, lit in enumerate(literals):
        if lit == "+" or lit == "-":
            processed_literals.append(lit)
        elif numerical_reg.match(lit):
            # might be a float.
            if "." in lit:
                num = float(lit)
            # Integer
            else:
                num = int(lit)
            # process to int if possible.
            if np.isclose(num, round(num)):
                num = round(num)
            processed_literals.append(num)
        else:
            sublattice_match = sublattice_reg.match(lit)
            # Specified a sub-lattice index.
            if sublattice_match is not None:
                species = get_species(sublattice_match.groups()[0])
                sublattice_id = int(sublattice_match.groups()[1])
            # No sub-lattice index.
            else:
                species = get_species(lit)
                sublattice_id = None
            processed_literals.append((species, sublattice_id))

    # Pack literals and intercept term.
    intercept = 0
    if isinstance(processed_literals[-1], Number):
        sym, n_sym = _extract_symbols_before_number(
            processed_literals, len(processed_literals) - 1
        )
        intercept = sym * processed_literals[-1]
        processed_literals = processed_literals[: -(1 + n_sym)]
    elif processed_literals[-1] == "+":
        intercept = 1
        processed_literals = processed_literals[:-1]
    elif processed_literals[-1] == "-":
        intercept = -1
        processed_literals = processed_literals[:-1]

    packed_literals = []
    # A species literal should never be preceded by more than 1 number literals.
    for i, lit in enumerate(processed_literals):
        if isinstance(lit, tuple):
            # Extract the number literal before a species literal.
            num, n_nums = _extract_number_before_species(processed_literals, i)
            sym, _ = _extract_symbols_before_number(processed_literals, i - n_nums)

            packed_literals.append((sym * num, *lit))

    return packed_literals, intercept


def convert_constraint_string(entry, bits):
    """Convert a string containing a composition constraint to standard form.

    Args:
        entry(str):
           A string containing a constraint to composition. For example:
               "2 Ag+(0) + Cl-(1) +3 H+(2) <= 3 Mn2+ +4"
           There are a few requirements for writing string:
           1, Three relations, "==", "<=", ">=", are allowed. These symbols
              separate the left and the right side. They must be wrapped by
              exactly one space before and one space after.
           2, You can include a number in brackets following a species
              string to constrain the amount of species in a particular
              sub-lattice. If no sub-lattice index is given, will apply
              the constraint to this particular species on all sub-lattices.
           3, Species strings (with sub-lattice index in the bracket, if any)
              must be separated from its preceding number coefficient by a
              single space.
           4, Species strings must be separated from other contents before
              and after them by at least one space, but no space is allowed
              within a species string, nor between a species string and its
              sub-lattice index specifier.
              For the format of species strings, refer to pymatgen.core.species.
           5, Intercept terms must be written at the end of each side of the
              equation.
           6, All coefficients and number terms in the equation must be set
              according to the sub-lattice sizes in CompositionSpace object.
        bits(list[list[Species|Vacancy|Element]]):
            Allowed species on each sub-lattice. Must be in the same ordering
            as will appear in moca.composition.
    Returns:
        list of Number, Number, str:
            Left side of the equation, right side of the equation (simplified
            to only contain a number), and the label of relation. All in "counts"
            format as specified in moca.composition.
    """
    entry = entry.strip()
    separator_reg = re.compile(r"^(.*) ([<=>]?=) (.*)$")
    sep = separator_reg.match(entry)
    if sep is None:
        raise ValueError(
            f"Provided string entry {entry} must have <=, >=, == or"
            " = symbol wrapped by two spaces in the middle!"
        )
    left_string, relation, right_string = sep.groups()

    # Separate left and right side into terms containing species or numbers.
    left_pack, left_intercept = handle_side_string(left_string)
    right_pack, right_intercept = handle_side_string(right_string)

    n_dims = sum([len(species) for species in bits])
    dim_ids = get_dim_ids_by_sublattice(bits)
    left_vec = [0 for _ in range(n_dims)]

    # Add left.
    for coef, spec, sl_id in left_pack:
        # Specified to constrain the species on only one sub-lattice.
        if sl_id is not None:
            if spec not in bits[sl_id]:
                raise ValueError(
                    f"Species {spec} cannot be found in the"
                    f" specified sub-lattice {sl_id}!"
                )
            dim_id = dim_ids[sl_id][bits[sl_id].index(spec)]
            left_vec[dim_id] += coef
        # The corresponding species on all sub-lattices should be constrained.
        else:
            if spec not in itertools.chain(*bits):
                raise ValueError(
                    f"Species {spec} cannot be found in any"
                    f" sub-lattice! Consider only species in"
                    f" {set(itertools.chain(*bits))}!"
                )
            for species, sub_dim_ids in zip(bits, dim_ids):
                if spec in species:
                    dim_id = sub_dim_ids[species.index(spec)]
                    left_vec[dim_id] += coef

    # Subtract right.
    for coef, spec, sl_id in right_pack:
        # Specified to constrain the species on only one sub-lattice.
        if sl_id is not None:
            if spec not in bits[sl_id]:
                raise ValueError(
                    f"Species {spec} cannot be found in the"
                    f" specified sub-lattice {sl_id}."
                    f" Consider species {bits[sl_id]}!"
                )
            dim_id = dim_ids[sl_id][bits[sl_id].index(spec)]
            left_vec[dim_id] -= coef
        # The corresponding species on all sub-lattices should be constrained.
        else:
            if spec not in itertools.chain(*bits):
                raise ValueError(
                    f"Species {spec} cannot be found in any"
                    f" sub-lattice! Consider only species in"
                    f" {set(itertools.chain(*bits))}!"
                )
            for species, sub_dim_ids in zip(bits, dim_ids):
                if spec in species:
                    dim_id = sub_dim_ids[species.index(spec)]
                    left_vec[dim_id] -= coef

    right = right_intercept - left_intercept

    return left_vec, right, relation


class CompositionConstraintsManager:
    """A descriptor class that manages setting composition constraints.

    Composition constraints can be provided in the string format or the
    dictionary format. See documentations in space.py.
    """

    def __set_name__(self, owner, name):
        """Set the private variable names."""
        self.public_name = name
        self.private_name = "_" + name

    def __get__(self, obj, objtype=None):
        """Return the chemical potentials if set None otherwise."""
        value = getattr(obj, self.private_name, None)
        return value if value is None else value["value"]

    def __set__(self, obj, value):
        """Set the table given the owner and value."""
        if value is None:  # call delete if set to None
            self.__delete__(obj)
            return

        # value must be list of tuples, each with a list and a number.
        # No scaling would be done. Take care when filling in!
        constraints_eq = []
        # All geq constraints will be rearranged into leq constraints.
        constraints_leq = []
        bits = obj.site_spaces

        for entry in value:
            if isinstance(entry, (tuple, list)):
                left_vec, right, relation = entry
            # Constraint set as a string.
            elif isinstance(entry, str):
                left_vec, right, relation = convert_constraint_string(entry, bits)
            else:
                raise ValueError(
                    f"Constraint {entry} format is not readable. See"
                    f" smol.moca.composition for detailed description of formats."
                )

            if relation == "leq" or relation == "<=":
                constraints_leq.append((left_vec, right))
            if relation == "geq" or relation == ">=":
                constraints_leq.append(([-1 * num for num in left_vec], -1 * right))
            if relation == "eq" or relation == "==" or relation == "=":
                constraints_eq.append((left_vec, right))

        value = {
            "leq": constraints_leq,
            "eq": constraints_eq,
        }

        # if first instantiation concatenate the natural parameter
        if not hasattr(obj, self.private_name):
            setattr(
                obj,
                self.private_name,
                {"value": value},
            )

    def __delete__(self, obj):
        """Delete the boundary condition."""
        if hasattr(obj, self.private_name):
            del obj.__dict__[self.private_name]



# Helpers to safely access OpenMP routines
#
# no-op implementations are provided for the case where OpenMP is not available.
#
# All calls to OpenMP routines should be cimported from this module.
#
# Code based on sklearn.utils._openmp_helpers

import os

# Module level cache for cpu_count as we do not expect this to change during
# the lifecycle of a Python program.
_CPU_COUNT: int = 1


cdef extern from *:
    """
    #ifdef _OPENMP
        #include <omp.h>
        #define OPENMP_ENABLED 1
    #else
        #define OPENMP_ENABLED 0
        #define omp_lock_t int
        #define omp_init_lock(l) (void)0
        #define omp_destroy_lock(l) (void)0
        #define omp_set_lock(l) (void)0
        #define omp_unset_lock(l) (void)0
        #define omp_get_thread_num() 0
        #define omp_get_max_threads() 1
    #endif
    """
    bint OPENMP_ENABLED

    int omp_get_thread_num() noexcept nogil
    int omp_get_max_threads() noexcept nogil


cpdef _openmp_effective_numthreads(n_threads=None):
    """Determine the effective number of threads to be used for OpenMP calls

    - For ``n_threads = None``,
      - if the ``OMP_NUM_THREADS`` environment variable is set, return
        ``openmp.omp_get_max_threads()``
      The result of ``omp_get_max_threads`` can be influenced by environment
      variable ``OMP_NUM_THREADS`` or at runtime by ``omp_set_num_threads``.

    - For ``n_threads > 0``, return this as the maximal number of threads for
      parallel OpenMP calls.

    - For ``n_threads < 0``, return the maximal number of threads minus
      ``|n_threads + 1|``. In particular ``n_threads = -1`` will use as many
      threads as there are available cores on the machine.

    - Raise a ValueError for ``n_threads = 0``.

    If smol is built without OpenMP support, always return 1.
    """
    if n_threads == 0:
        raise ValueError("n_threads = 0 is invalid")

    if not OPENMP_ENABLED:
        # OpenMP disabled at build-time => sequential mode
        return 1

    if os.getenv("OMP_NUM_THREADS"):
        # Fall back to user provided number of threads making it possible
        # to exceed the number of cpus.
        max_n_threads = omp_get_max_threads()
    else:
        max_n_threads = min(omp_get_max_threads(), os.cpu_count())

    if n_threads is None:
        return max_n_threads
    elif n_threads < 0:
        return max(1, max_n_threads + n_threads + 1)

    return n_threads


cpdef _openmp_enabled():
    """Return true if openmp is available and enabled in build."""
    return OPENMP_ENABLED == 1


cpdef _openmp_get_max_threads():
    """Return the number of threads available to the OpenMP runtime."""
    return omp_get_max_threads()



"""
Class implementations for ClusterExpansion based Monte Carlo.

The moca (pronounced mocha) holds implementations of classes used to run
Monte Carlo simulations using Cluster Expansion Hamiltonians.
"""

from smol.moca.composition.space import CompositionSpace
from smol.moca.ensemble import Ensemble
from smol.moca.kernel import available_mckernels
from smol.moca.kernel.bias import available_bias_types
from smol.moca.kernel.mcusher import available_step_types
from smol.moca.processor.composite import CompositeProcessor
from smol.moca.processor.ewald import EwaldProcessor
from smol.moca.processor.expansion import (
    ClusterDecompositionProcessor,
    ClusterExpansionProcessor,
)
from smol.moca.sampler.container import SampleContainer
from smol.moca.sampler.sampler import Sampler

__all__ = [
    "ClusterExpansionProcessor",
    "ClusterDecompositionProcessor",
    "EwaldProcessor",
    "CompositeProcessor",
    "Ensemble",
    "Sampler",
    "SampleContainer",
    "CompositionSpace",
    "available_mckernels",
    "available_step_types",
    "available_bias_types",
]



import os

import numpy as np
import pytest
from monty.serialization import loadfn
from pymatgen.core import Structure

from smol.cofe import ClusterExpansion, ClusterSubspace, StructureWrangler
from smol.cofe.extern import EwaldTerm
from smol.cofe.space.basis import BasisIterator
from smol.moca import Ensemble
from smol.moca.processor import (
    ClusterDecompositionProcessor,
    ClusterExpansionProcessor,
    CompositeProcessor,
    EwaldProcessor,
)
from smol.moca.processor.distance import (
    ClusterInteractionDistanceProcessor,
    CorrelationDistanceProcessor,
)
from smol.utils.class_utils import get_subclasses
from tests.utils import gen_fake_training_data

# uncomment below to show HDF5 C traceback
# import h5py
# h5py._errors.unsilence_errors()


SEED = None

# load test data files and set them up as fixtures
DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

# some test structures and other parameters for creating fixtures
files = [
    "AuPd_prim.json",
    "CrFeW_prim.json",
    "LiCaBr_prim.json",
    "LiMOF_prim.json",
    "LiMnTiVOF_prim.json",
]

test_structures = [loadfn(os.path.join(DATA_DIR, file)) for file in files]
basis_iterator_names = list(get_subclasses(BasisIterator))


@pytest.fixture(scope="module")
def rng():
    """Seed and return an RNG for test reproducibility"""
    return np.random.default_rng(SEED)


@pytest.fixture(scope="package")
def cluster_cutoffs():
    # parametrize this if we ever want to test on different cutoffs
    return {2: 6, 3: 5, 4: 4}


@pytest.fixture(params=test_structures, scope="package")
def structure(request):
    return request.param


@pytest.fixture(scope="package")
def expansion_structure(structure):
    sites = [
        site
        for site in structure
        if site.species.num_atoms < 0.99 or len(site.species) > 1
    ]
    return Structure.from_sites(sites)


@pytest.fixture(scope="package")
def single_structure():
    # this is the LiCaBr structure used for some fixed tests
    return test_structures[2]


@pytest.fixture(params=test_structures, scope="package")
def cluster_subspace(cluster_cutoffs, request):
    subspace = ClusterSubspace.from_cutoffs(
        request.param, cutoffs=cluster_cutoffs, supercell_size="volume"
    )
    return subspace


@pytest.fixture(params=test_structures, scope="package")
def cluster_subspace_ewald(cluster_cutoffs, request):
    subspace = ClusterSubspace.from_cutoffs(
        request.param, cutoffs=cluster_cutoffs, supercell_size="volume"
    )
    subspace.add_external_term(EwaldTerm())
    return subspace


@pytest.fixture(scope="package")
def single_subspace(single_structure):
    # this is a subspace with the LiCaBr structure for some fixed tests
    subspace = ClusterSubspace.from_cutoffs(
        single_structure, cutoffs={2: 6, 3: 5}, supercell_size="volume"
    )
    return subspace


# fixture for all processors
@pytest.fixture(
    params=[
        "expansion",
        "decomposition",
        "ewald",
        "composite",
        "corr_distance",
        "int_distance",
    ],
    scope="module",
)
def processor(cluster_subspace, rng, request):
    coefs = 2 * np.random.random(cluster_subspace.num_corr_functions)
    scmatrix = 3 * np.eye(3)

    if request.param == "expansion":
        proc = ClusterExpansionProcessor(cluster_subspace, scmatrix, coefficients=coefs)
    elif request.param == "decomposition":
        expansion = ClusterExpansion(cluster_subspace, coefs)
        proc = ClusterDecompositionProcessor(
            cluster_subspace, scmatrix, expansion.cluster_interaction_tensors
        )
    elif request.param == "ewald":
        proc = EwaldProcessor(cluster_subspace, scmatrix, EwaldTerm(), coefficient=1.0)
    elif request.param == "composite":
        proc = CompositeProcessor(cluster_subspace, supercell_matrix=scmatrix)
        proc.add_processor(
            ClusterExpansionProcessor(cluster_subspace, scmatrix, coefficients=coefs)
        )
        proc.add_processor(
            EwaldProcessor(cluster_subspace, scmatrix, EwaldTerm(), coefficient=1.0)
        )
    elif request.param == "corr_distance":
        proc = CorrelationDistanceProcessor(cluster_subspace, scmatrix)
    else:
        expansion = ClusterExpansion(cluster_subspace, coefs)
        proc = ClusterInteractionDistanceProcessor(
            cluster_subspace, scmatrix, expansion.cluster_interaction_tensors
        )

    yield proc
    cluster_subspace._external_terms = []  # Ewald processor will add one..


@pytest.fixture(scope="module")
def ce_processor(cluster_subspace, rng):
    coefs = 2 * rng.random(cluster_subspace.num_corr_functions)
    scmatrix = 3 * np.eye(3)
    return ClusterExpansionProcessor(
        cluster_subspace, supercell_matrix=scmatrix, coefficients=coefs
    )


@pytest.fixture(params=["expansion", "decomposition"], scope="module")
def composite_processor(cluster_subspace_ewald, rng, request):
    coefs = 2 * np.random.random(cluster_subspace_ewald.num_corr_functions + 1)
    scmatrix = 3 * np.eye(3)
    proc = CompositeProcessor(cluster_subspace_ewald, supercell_matrix=scmatrix)
    if request.param == "expansion":
        proc.add_processor(
            ClusterExpansionProcessor(
                cluster_subspace_ewald, scmatrix, coefficients=coefs[:-1]
            )
        )
    else:  # elif request.param == "decomposition":
        expansion = ClusterExpansion(cluster_subspace_ewald, coefs)
        proc.add_processor(
            ClusterDecompositionProcessor(
                cluster_subspace_ewald, scmatrix, expansion.cluster_interaction_tensors
            )
        )

    proc.add_processor(
        EwaldProcessor(
            cluster_subspace_ewald, scmatrix, EwaldTerm(), coefficient=coefs[-1]
        )
    )
    # bind raw coefficients since OD processors do not store them
    # and be able to test computing properties, hacky but oh well
    proc.raw_coefs = coefs
    return proc


@pytest.fixture(params=["canonical", "semigrand"], scope="module")
def ensemble(composite_processor, request):
    if request.param == "semigrand":
        species = {
            sp
            for space in composite_processor.active_site_spaces
            for sp in space.keys()
        }
        kwargs = {"chemical_potentials": {sp: 0.3 for sp in species}}
    else:
        kwargs = {}
    return Ensemble(composite_processor, **kwargs)


@pytest.fixture(scope="module")
def single_sgc_ensemble(rng):
    # a single sgc ensemble using the LMOF test structures
    subspace = ClusterSubspace.from_cutoffs(
        test_structures[3], cutoffs={2: 6, 3: 4}, supercell_size="volume"
    )
    coefs = rng.random(subspace.num_corr_functions)
    coefs[0] = -1.0
    proc = ClusterExpansionProcessor(subspace, 6 * np.eye(3), coefs)
    species = {sp for space in proc.active_site_spaces for sp in space.keys()}
    return Ensemble(proc, chemical_potentials={sp: 1.0 for sp in species})


@pytest.fixture(scope="module")
def single_canonical_ensemble(single_subspace, rng):
    coefs = rng.random(single_subspace.num_corr_functions)
    proc = ClusterExpansionProcessor(single_subspace, 4 * np.eye(3), coefs)
    return Ensemble(proc)


@pytest.fixture(params=basis_iterator_names, scope="package")
def basis_name(request):
    return request.param.split("Iterator")[0]


@pytest.fixture
def supercell_matrix(rng):
    def skewed(scm):  # check if any angle is less than 45 degrees
        normed_scm = scm / np.linalg.norm(scm, axis=0)
        off_diag = (np.array([0, 0, 1]), np.array([1, 2, 2]))
        return any(abs(normed_scm @ normed_scm)[off_diag] > 0.5)

    # make sure not singular and not overly skewed
    while abs(np.linalg.det(m := rng.integers(-3, 3, size=(3, 3)))) < 1e-4 or skewed(m):
        pass

    return m


@pytest.fixture
def structure_wrangler(single_subspace, rng):
    wrangler = StructureWrangler(single_subspace)
    for entry in gen_fake_training_data(single_subspace.structure, n=10, rng=rng):
        wrangler.add_entry(entry, weights={"random": 2.0})
    yield wrangler
    # force remove any external terms added in tests
    wrangler.cluster_subspace._external_terms = []



"""Test constraints to the upper-bound problem."""

import itertools

import pytest
from pymatgen.core import Species

from smol.capp.generate.groundstate.upper_bound.constraints import (
    get_composition_space_constraints,
    get_fixed_composition_constraints,
    get_normalization_constraints,
)
from smol.capp.generate.groundstate.upper_bound.variables import (
    get_variables_from_sublattices,
)
from smol.moca.occu_utils import get_dim_ids_table, occu_to_counts

from .utils import get_random_neutral_variable_values, get_random_variable_values


def test_normalization(solver_test_ensemble, solver_test_initial_occupancy):
    variables, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )
    constraints = get_normalization_constraints(variables, variable_indices)
    assert len(constraints) == solver_test_ensemble.num_sites - len(
        solver_test_ensemble.restricted_sites
    )
    # Check with constraint compliant random occupancies.
    for _ in range(20):
        rand_val = get_random_variable_values(solver_test_ensemble.sublattices)
        variables.value = rand_val
        for c in constraints:
            assert c.value()


def test_comp_space_constraints(solver_test_ensemble, solver_test_initial_occupancy):
    variables, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )
    constraints = get_composition_space_constraints(
        solver_test_ensemble.sublattices,
        variables,
        variable_indices,
        solver_test_ensemble.processor.structure,
    )

    # Only 1 charge balance because normalization skipped.
    assert len(constraints) == 1
    for _ in range(20):
        rand_val = get_random_neutral_variable_values(
            solver_test_ensemble.sublattices,
            solver_test_initial_occupancy,
            variable_indices,
        )
        variables.value = rand_val
        for c in constraints:
            assert c.value()

    # Add one more constraint to constrain some species amounts.
    constraints = get_composition_space_constraints(
        solver_test_ensemble.sublattices,
        variables,
        variable_indices,
        solver_test_ensemble.processor.structure,
        other_constraints=[
            "Mn4+ == 2",  # Broken when force_flip, kept when canonical. 2nd.
            "Mn4+ + Mn2+ >= 7",  # Never true. 3rd.
            "Mn4+ + Mn2+ <= 6",  # Always true. 4th.
            "0 >= -1",  # Always True. Skipped.
            "0 <= 1.5",  # Always True. Skipped.
            "0.0 = 0.0",  # Always True. Skipped.
        ],
    )

    assert len(constraints) == 4
    # Check with force_flip.
    for _ in range(20):
        rand_val = get_random_neutral_variable_values(
            solver_test_ensemble.sublattices,
            solver_test_initial_occupancy,
            variable_indices,
            force_flip=True,
        )
        variables.value = rand_val
        results = [c.value() for c in constraints]
        assert results == [True, False, False, True]
    # Check with canonical.
    for _ in range(20):
        rand_val = get_random_neutral_variable_values(
            solver_test_ensemble.sublattices,
            solver_test_initial_occupancy,
            variable_indices,
            canonical=True,
        )
        variables.value = rand_val
        results = [c.value() for c in constraints]
        assert results == [True, True, False, True]

    # Bad test cases.
    with pytest.raises(ValueError):
        _ = get_composition_space_constraints(
            solver_test_ensemble.sublattices,
            variables,
            variable_indices,
            solver_test_ensemble.processor.structure,
            other_constraints=[
                " == 10",  # Unsatisfiable.
            ],
        )
    with pytest.raises(ValueError):
        _ = get_composition_space_constraints(
            solver_test_ensemble.sublattices,
            variables,
            variable_indices,
            solver_test_ensemble.processor.structure,
            other_constraints=[
                " >= 1",  # Unsatisfiable.
            ],
        )
    with pytest.raises(ValueError):
        _ = get_composition_space_constraints(
            solver_test_ensemble.sublattices,
            variables,
            variable_indices,
            solver_test_ensemble.processor.structure,
            other_constraints=[
                " <= -1",  # Unsatisfiable.
            ],
        )
    with pytest.raises(ValueError):
        _ = get_composition_space_constraints(
            solver_test_ensemble.sublattices,
            variables,
            variable_indices,
            solver_test_ensemble.processor.structure,
            other_constraints=[
                "F- == 1000",  # Unsatisfiable.
            ],
        )
    with pytest.raises(ValueError):
        _ = get_composition_space_constraints(
            solver_test_ensemble.sublattices,
            variables,
            variable_indices,
            solver_test_ensemble.processor.structure,
            other_constraints=[
                "F- >= 1000",  # Unsatisfiable.
            ],
        )
    with pytest.raises(ValueError):
        _ = get_composition_space_constraints(
            solver_test_ensemble.sublattices,
            variables,
            variable_indices,
            solver_test_ensemble.processor.structure,
            other_constraints=[
                "F- <= 1",  # Unsatisfiable.
            ],
        )


def test_fixed_composition_constraints(
    solver_test_ensemble, solver_test_initial_occupancy
):
    variables, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )
    bits = [s.species for s in solver_test_ensemble.sublattices]
    n_dims = sum([len(sl_species) for sl_species in bits])
    table = get_dim_ids_table(solver_test_ensemble.sublattices)
    fixed_counts = occu_to_counts(solver_test_initial_occupancy, n_dims, table)
    constraints = get_fixed_composition_constraints(
        solver_test_ensemble.sublattices,
        variables,
        variable_indices,
        solver_test_ensemble.processor.structure,
        fixed_composition=fixed_counts,
    )

    # F- and O2- are fixed and always satisfied, will not appear.
    assert len(constraints) == 5
    for _ in range(20):
        rand_val = get_random_neutral_variable_values(
            solver_test_ensemble.sublattices,
            solver_test_initial_occupancy,
            variable_indices,
            canonical=True,
        )  # Force canonical constraints, will always satisfy.
        variables.value = rand_val
        results = [c.value() for c in constraints]
        assert results == [True for _ in range(5)]
    flatten_bits = list(itertools.chain(*bits))
    flatten_bits.remove(Species("F", -1))
    flatten_bits.remove(Species("O", -2))
    assert len(flatten_bits) == 5
    ti_id = flatten_bits.index(Species("Ti", 4))
    mn4_id = flatten_bits.index(Species("Mn", 4))
    for _ in range(20):
        rand_val = get_random_neutral_variable_values(
            solver_test_ensemble.sublattices,
            solver_test_initial_occupancy,
            variable_indices,
            force_flip=True,
        )  # Force Ti-Mn4 flip, will not satisfy Ti, Mn constraints
        variables.value = rand_val
        results = [c.value() for c in constraints]
        assumed_results = [
            (False if i in [ti_id, mn4_id] else True) for i in range(len(flatten_bits))
        ]
        assert results == assumed_results



"""Boilerplate code for a progress bar using tqdm or no bar if not installed."""

import warnings


def progress_bar(display, total, description):
    """Get a tqdm progress bar interface.

    If the tqdm library is not installed, this will be an empty progress bar
    that does nothing.

    Args:
        display (bool):
            if true, a real progress bar will be returned.
        total (int):
            the total size of the progress bar.
        description (str):
            description to print in progress bar.
    """
    try:
        # pylint: disable=import-outside-toplevel
        import tqdm
    except ImportError:
        tqdm = None

    if display:
        if tqdm is None:
            warnings.warn(
                "tqdm library needs to be installed to show a " " progress bar."
            )
            return _EmptyBar()

        return tqdm.tqdm(total=total, desc=description)
        # if display is True:
        #   return tqdm.tqdm(total=total, desc=description)
        # else:
        #    return getattr(tqdm, "tqdm_" + display)(total=total)
    return _EmptyBar()


class _EmptyBar:
    """A dummy progress bar.

    Idea take from emce:
    https://github.com/dfm/emcee/blob/main/src/emcee/pbar.py
    """

    # pylint: disable=missing-function-docstring

    def __init__(self):
        pass

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def update(self, *args):
        pass



"""Implementation of CE processor class for a fixed size supercell.

If you are using a Hamiltonian with an Ewald summation electrostatic term, you
should use the CompositeProcessor with a ClusterExpansionProcessor and an
EwaldProcessor class to handle changes in the electrostatic interaction energy.
"""

__author__ = "Luis Barroso-Luque"

from collections import defaultdict
from dataclasses import dataclass, field

import numpy as np

from smol.cofe.space.clusterspace import ClusterSubspace, OrbitIndices
from smol.moca.processor.base import Processor
from smol.utils.cluster import get_orbit_data
from smol.utils.cluster.container import IntArray2DContainer
from smol.utils.cluster.evaluator import ClusterSpaceEvaluator
from smol.utils.cluster.numthreads import SetNumThreads
from smol.utils.setmany import SetMany


@dataclass
class LocalEvalData:
    """Holds data for local evaluation updates.

    A dataclass to the data for local evaluation updates of correlation and cluster
    interaction vectors.
    """

    site_index: int
    evaluator: ClusterSpaceEvaluator = field(repr=False)
    indices: OrbitIndices = field(repr=False)
    cluster_ratio: np.ndarray
    num_threads: int = field(default=SetNumThreads("evaluator"), repr=True)


class ClusterExpansionProcessor(Processor):
    """ClusterExpansionProcessor class to use a ClusterExpansion in MC.

    A CE processor is optimized to compute correlation vectors and local
    changes in correlation vectors. This class allows the use of a CE
    Hamiltonian to run Monte Carlo-based simulations

    A processor that allows an Ensemble class to generate a Markov chain
    for sampling thermodynamic properties from a CE Hamiltonian.

    Attributes:
        coefs (ndarray):
            Fitted coefficients from the cluster expansion.
        num_corr_functions (int):
            Total number of orbit basis functions (correlation functions).
            This includes all possible labelings/orderings for all orbits.
            Same as :code:`ClusterSubspace.n_bit_orderings`.
    """

    num_threads = SetMany("num_threads", "_eval_data_by_sites")
    num_threads_full = SetNumThreads("_evaluator")

    def __init__(
        self,
        cluster_subspace,
        supercell_matrix,
        coefficients,
        num_threads=None,
        num_threads_full=None,
    ):
        """Initialize a ClusterExpansionProcessor.

        Args:
            cluster_subspace (ClusterSubspace):
                a cluster subspace
            supercell_matrix (ndarray):
                an array representing the supercell matrix with respect to the
                Cluster Expansion prim structure.
            coefficients (ndarray):
                fit coefficients for the represented cluster expansion.
            num_threads (int): optional
                Number of threads to use to compute `updates in a correlation vector
                from local changes. If None is given the number of threads are set
                to the same as set in the given :class:`ClusterSubspace`.
                Note that this is not saved when serializing the ClusterSubspace with
                the as_dict method, so if you are loading a ClusterSubspace from a
                file then make sure to set the number of threads as desired.
            num_threads_full (int): optional
                Number of threads to use to compute a full correlation vector. Note
                that this is not saved when serializing the ClusterSubspace with the
                as_dict method, so if you are loading a ClusterSubspace from a
                file then make sure to set the number of threads as desired.
        """
        super().__init__(cluster_subspace, supercell_matrix, coefficients)

        if len(coefficients) != self.cluster_subspace.num_corr_functions:
            raise ValueError(
                f"The provided coefficients are not the right length. "
                f"Got {len(coefficients)} coefficients, the length must be "
                f"{self.cluster_subspace.num_corr_functions} based on the provided cluster "
                f"subspace."
            )

        # evaluator for the correlation functions
        self._evaluator = ClusterSpaceEvaluator(
            get_orbit_data(cluster_subspace.orbits),
            cluster_subspace.num_orbits,
            cluster_subspace.num_corr_functions,
        )

        # orbit indices mapping all sites in this supercell to clusters in each
        # orbit. This is used to compute the correlation vector.
        self._indices = cluster_subspace.get_orbit_indices(supercell_matrix)

        # TODO there is a lot of repeated code here and in the CD __init__
        # TODO we should probably refactor this to avoid this and make testing easier
        # Dictionary of orbits data by site index necessary to compute local changes in
        # correlation vectors from flips
        data_by_sites = defaultdict(list)
        # We also store a reduced cluster index array where only the rows with the site
        # index are stored. The ratio is also needed because the correlations are
        # averages over the full cluster indices array.
        for orbit, cluster_indices in zip(
            cluster_subspace.orbits, self._indices.arrays
        ):
            orbit_data = (
                orbit.id,
                orbit.bit_id,
                orbit.flat_correlation_tensors,
                orbit.flat_tensor_indices,
            )
            for site_ind in np.unique(cluster_indices):
                in_inds = np.any(cluster_indices == site_ind, axis=-1)
                ratio = len(cluster_indices) / np.sum(in_inds)
                data_by_sites[site_ind].append(
                    (orbit_data, cluster_indices[in_inds], ratio)
                )

        # now unpack the data, create the local evaluators and store them in a
        # dictionary indexed by site index.
        self._eval_data_by_sites = {}
        for site, data in data_by_sites.items():
            # we don't really need a different evaluator for each site since sym
            # equivalent sites can share the same one, same goes for ratio.
            evaluator = ClusterSpaceEvaluator(
                tuple(d[0] for d in data),
                cluster_subspace.num_orbits,
                cluster_subspace.num_corr_functions,
            )
            indices = tuple(d[1] for d in data)
            orbit_indices = OrbitIndices(indices, IntArray2DContainer(indices))
            ratio = np.array([d[2] for d in data])
            self._eval_data_by_sites[site] = LocalEvalData(
                site, evaluator, orbit_indices, ratio, 1
            )

        # Set the number of threads to use for the full correlation vector
        self.num_threads_full = (
            num_threads_full if num_threads_full else cluster_subspace.num_threads
        )
        # Set the number of threads to use for the local correlation vector
        self.num_threads = num_threads

    def compute_feature_vector(self, occupancy):
        """Compute the correlation vector for a given occupancy string.

        The correlation vector in this case is normalized per supercell. In
        other works it is extensive corresponding to the size of the supercell.

        Args:
            occupancy (ndarray[np.int32]):
                encoded occupation array

        Returns:
            array: correlation vector
        """
        try:
            corr = (
                self._evaluator.correlations_from_occupancy(
                    occupancy, self._indices.container
                )
                * self.size
            )
        except ValueError:
            raise ValueError(
                f"occupancy dtype is: {occupancy.dtype}, must be an array of np.int32!"
            )
        return corr

    def compute_feature_vector_change(self, occupancy, flips):
        """
        Compute the change in the correlation vector from a list of flips.

        The correlation vector in this case is normalized per supercell. In
        other works it is extensive corresponding to the size of the supercell.

        Args:
            occupancy (ndarray):
                encoded occupancy string
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a
                single flip where the first element is the index of the site
                in the occupancy string and the second element is the index
                for the new species to place at that site.

        Returns:
            array: change in correlation vector
        """
        occu_i = occupancy
        delta_corr = np.zeros(self.cluster_subspace.num_corr_functions)
        for f in flips:
            occu_f = occu_i.copy()
            occu_f[f[0]] = f[1]
            eval_data = self._eval_data_by_sites[f[0]]
            try:
                delta_corr += eval_data.evaluator.delta_correlations_from_occupancies(
                    occu_f, occu_i, eval_data.cluster_ratio, eval_data.indices.container
                )
            except ValueError:
                raise ValueError(
                    f"occupancy dtype is: {occupancy.dtype}, must be an array of np.int32!"
                )
            occu_i = occu_f

        return delta_corr * self.size

    @classmethod
    def from_dict(cls, d):
        """Create a ClusterExpansionProcessor from serialized MSONable dict."""
        return cls(
            ClusterSubspace.from_dict(d["cluster_subspace"]),
            np.array(d["supercell_matrix"]),
            coefficients=np.array(d["coefficients"]),
        )


class ClusterDecompositionProcessor(Processor):
    """ClusterDecompositionProcessor to use a CD in MC simulations.

    An ClusterDecompositionProcessor is similar to a ClusterExpansionProcessor
    however it computes properties and changes of properties using cluster interactions
    rather than correlation functions and ECI.

    For a given ClusterExpansion the results from sampling using this class
    should be the exact same as using a ClusterExpansionProcessor (ie the
    samples are from the same ensemble).

    A few practical and important differences though, the featyre vectors
    that are used and sampled are the cluster interactions for each occupancy
    as opposed to the correlation vectors (so if you need correlation vectors
    for further analysis you're probably better off with a standard
    ClusterExpansionProcessor

    However, if you don't need correlation vectors, then this is probably
    more efficient and more importantly the scaling complexity with model size
    is in terms of number of orbits (instead of number of corr functions).
    This can give substantial speedups when doing MC calculations of complex
    high component systems.
    """

    num_threads = SetMany("num_threads", "_eval_data_by_sites")
    num_threads_full = SetNumThreads("_evaluator")

    def __init__(
        self,
        cluster_subspace,
        supercell_matrix,
        interaction_tensors,
        coefficients=None,
        num_threads=None,
        num_threads_full=None,
    ):
        """Initialize an ClusterDecompositionProcessor.

        Args:
            cluster_subspace (ClusterSubspace):
                A cluster subspace
            supercell_matrix (ndarray):
                An array representing the supercell matrix with respect to the
                Cluster Expansion prim structure.
            interaction_tensors (sequence of ndarray):
                Sequence of ndarray where each array corresponds to the
                cluster interaction tensor. These should be in the same order as their
                corresponding orbits in the given cluster_subspace.
            num_threads (int): optional
                Number of threads to use to compute a correlation vector. Note that
                this is not saved when serializing the ClusterSubspace with the
                as_dict method, so if you are loading a ClusterSubspace from a
                file then make sure to set the number of threads as desired.
                corresponding orbits in the given cluster_subspace
            coefficients (ndarray): optional
                coefficients of each mean cluster interaction tensor. This should often
                be some multiple of the orbit multiplicities. Default is the orbit
                multiplicities.
        """
        if len(interaction_tensors) != cluster_subspace.num_orbits:
            raise ValueError(
                f"The number of cluster interaction tensors must match the number "
                f" of orbits in the subspace. Got {len(interaction_tensors)} "
                f"interaction tensors, but need {cluster_subspace.num_orbits} "
                f" for the given cluster_subspace."
            )

        #  the orbit multiplicities play the role of coefficients here.
        coefficients = (
            cluster_subspace.orbit_multiplicities
            if coefficients is None
            else coefficients
        )
        super().__init__(cluster_subspace, supercell_matrix, coefficients=coefficients)

        self._interaction_tensors = interaction_tensors  # keep these for serialization

        flat_interaction_tensors = tuple(
            np.ravel(tensor, order="C") for tensor in interaction_tensors[1:]
        )

        self._evaluator = ClusterSpaceEvaluator(
            get_orbit_data(cluster_subspace.orbits),
            cluster_subspace.num_orbits,
            cluster_subspace.num_corr_functions,
            1,  # set threads to here so we can set it later
            interaction_tensors[0],
            flat_interaction_tensors,
        )

        # orbit indices mapping all sites in this supercell to clusters in each
        # orbit. This is used to compute the correlation vector.
        self._indices = cluster_subspace.get_orbit_indices(supercell_matrix)

        # Dictionary of orbits data by site index necessary to compute local changes in
        # correlation vectors from flips
        data_by_sites = defaultdict(list)
        # We also store a reduced cluster index array where only the rows with the site
        # index are stored. The ratio is also needed because the correlations are
        # averages over the full cluster indices array.
        for orbit, cluster_indices, interaction_tensor in zip(
            cluster_subspace.orbits, self._indices.arrays, flat_interaction_tensors
        ):
            orbit_data = (
                orbit.id,
                orbit.bit_id,
                orbit.flat_correlation_tensors,
                orbit.flat_tensor_indices,
            )
            for site_ind in np.unique(cluster_indices):
                in_inds = np.any(cluster_indices == site_ind, axis=-1)
                ratio = len(cluster_indices) / np.sum(in_inds)
                data_by_sites[site_ind].append(
                    (orbit_data, cluster_indices[in_inds], ratio, interaction_tensor)
                )

        # now unpack the data, create the local evaluators and store them in a
        # dictionary indexed by site index.
        self._eval_data_by_sites = {}
        for site, data in data_by_sites.items():
            orbit_data = tuple(d[0] for d in data)
            indices = tuple(d[1] for d in data)
            orbit_indices = OrbitIndices(indices, IntArray2DContainer(indices))
            ratio = np.array([d[2] for d in data])
            flat_interaction_tensors = tuple(d[3] for d in data)
            # we don't really need a different evaluator for each site since sym
            # equivalent sites can share the same one, same goes for ratio.
            evaluator = ClusterSpaceEvaluator(
                orbit_data,
                cluster_subspace.num_orbits,
                cluster_subspace.num_corr_functions,
                1,  # set to single thread, will be set by num_threads later
                interaction_tensors[0],
                flat_interaction_tensors,
            )
            self._eval_data_by_sites[site] = LocalEvalData(
                site, evaluator, orbit_indices, ratio, 1
            )

        # Set the number of threads to use for the full correlation vector
        self.num_threads_full = (
            num_threads_full if num_threads_full else cluster_subspace.num_threads
        )
        # Set the number of threads to use for the local correlation vector
        self.num_threads = num_threads

    def compute_feature_vector(self, occupancy):
        """Compute the cluster interaction vector for a given occupancy string.

        The cluster interaction vector in this case is normalized per supercell.

        Args:
            occupancy (ndarray):
                encoded occupation array

        Returns:
            array: correlation vector
        """
        try:
            corr = (
                self._evaluator.interactions_from_occupancy(
                    occupancy, self._indices.container
                )
                * self.size
            )
        except ValueError:
            raise ValueError(
                f"occupancy dtype is: {occupancy.dtype}, must be an array of np.int32!"
            )
        return corr

    def compute_feature_vector_change(self, occupancy, flips):
        """
        Compute the change in the cluster interaction vector from a list of flips.

        The cluster interaction vector in this case is normalized per supercell.

        Args:
            occupancy (ndarray):
                encoded occupancy string
            flips (list of tuple):
                list of tuples with two elements. Each tuple represents a
                single flip where the first element is the index of the site
                in the occupancy string and the second element is the index
                for the new species to place at that site.

        Returns:
            array: change in cluster interaction vector
        """
        occu_i = occupancy
        delta_interactions = np.zeros(self.cluster_subspace.num_orbits)
        for f in flips:
            occu_f = occu_i.copy()
            occu_f[f[0]] = f[1]
            eval_data = self._eval_data_by_sites[f[0]]
            try:
                delta_interactions += (
                    eval_data.evaluator.delta_interactions_from_occupancies(
                        occu_f,
                        occu_i,
                        eval_data.cluster_ratio,
                        eval_data.indices.container,
                    )
                )
            except ValueError:
                raise ValueError(
                    f"occupancy dtype is: {occupancy.dtype}, must be an array of np.int32!"
                )
            occu_i = occu_f

        return delta_interactions * self.size

    def as_dict(self) -> dict:
        """
        Json-serialization dict representation.

        Returns:
            MSONable dict
        """
        d = super().as_dict()
        d["interaction_tensors"] = [
            interaction.tolist() for interaction in self._interaction_tensors
        ]
        return d

    @classmethod
    def from_dict(cls, d):
        """Create a ClusterExpansionProcessor from serialized MSONable dict."""
        interaction_tensors = tuple(
            np.array(interaction) for interaction in d["interaction_tensors"]
        )
        return cls(
            ClusterSubspace.from_dict(d["cluster_subspace"]),
            np.array(d["supercell_matrix"]),
            interaction_tensors,
        )



"""Sample Trace objects as simple namespaces."""

from types import SimpleNamespace

import numpy as np


class Trace(SimpleNamespace):
    """Simple Trace class.

    A Trace is a simple namespace to hold states and values to be recorded
    during MC sampling.
    """

    def __init__(self, /, **kwargs):  # noqa
        if not all(isinstance(val, np.ndarray) for val in kwargs.values()):
            raise TypeError("Trace only supports attributes of type ndarray.")
        super().__init__(**kwargs)

    @property
    def names(self):
        """Get all attribute names."""
        return tuple(self.__dict__.keys())

    def items(self):
        """Return generator for (name, attribute)."""
        yield from self.__dict__.items()

    def __setattr__(self, name, value):
        """Set only ndarrays as attributes."""
        if isinstance(value, float):
            value = np.array([value], dtype=np.float64)
        if isinstance(value, int):
            value = np.array([value], dtype=np.int32)

        # Don't enforce int32 and float64 here as that may cause overhead.
        # Calling modules will be responsible for checking dtype.
        if not isinstance(value, np.ndarray):
            raise TypeError("Trace only supports attributes of type ndarray.")

        self.__dict__[name] = value

    def as_dict(self):
        """Return copy of underlying dictionary."""
        return self.__dict__.copy()


class StepTrace(Trace):
    """StepTrace class.

    Same as Trace above but holds a default "delta_trace" inner trace to hold
    trace values that represent changes from previous values, to be handled
    similarly to delta_features and delta_energy.

    A StepTrace object is set as an MCKernel's attribute to record
    kernel specific values during sampling.
    """

    def __init__(self, /, **kwargs):  # noqa
        super().__init__(**kwargs)
        super(Trace, self).__setattr__("delta_trace", Trace())

    @property
    def names(self):
        """Get all field names. Removes delta_trace from field names."""
        return tuple(name for name in super().names if name != "delta_trace")

    def items(self):
        """Return generator for (name, attribute). Skips delta_trace."""
        for name, value in self.__dict__.items():
            if name == "delta_trace":
                continue
            yield name, value

    def __setattr__(self, name, value):
        """Set only ndarrays as attributes."""
        if name == "delta_trace":
            raise ValueError("Attribute name 'delta_trace' is reserved.")
        if not isinstance(value, np.ndarray):
            raise TypeError("Trace only supports attributes of type ndarray.")
        self.__dict__[name] = value

    def as_dict(self):
        """Return copy of serializable dictionary."""
        step_trace_d = self.__dict__.copy()
        step_trace_d["delta_trace"] = step_trace_d["delta_trace"].as_dict()
        return step_trace_d



"""Handles boolean product terms in expression."""

from typing import List, Tuple

import cvxpy as cp
import numpy as np
from cvxpy.constraints.constraint import Constraint
from numpy.typing import ArrayLike


# TODO: in the future, allow writing ground-state input files from terms,
#  which may support non-linear objectives and constraints.
def get_auxiliary_variable_values(
    variable_values: ArrayLike, indices_in_auxiliary_products: List[List[int]]
) -> np.ndarray:
    """Get the value of auxiliary variables from site variables.

    Args:
        variable_values(ArrayLike of bool or int):
            Values of site variables.
        indices_in_auxiliary_products(list of lists of int):
            A list containing the indices of variables whose product equals to the
            corresponding auxiliary slack variable.
    Returns:
        np.ndarray:
            Values of auxiliary variables subjecting to auxiliary constraints,
            as 0 and 1.
    """
    variable_values = np.array(variable_values, dtype=int)
    aux_values = np.ones(len(indices_in_auxiliary_products), dtype=int)
    for i, inds in enumerate(indices_in_auxiliary_products):
        aux_values[i] = np.prod(variable_values[inds])

    return aux_values.astype(int)


def get_expression_and_auxiliary_from_terms(
    cluster_terms: List[Tuple[List[int], float, float]],
    variables: cp.Variable,
    coefficients_cutoff: float = 0.0,
) -> Tuple[cp.Expression, cp.Variable, List[List[int]], List[Constraint]]:
    """Convert the cluster terms into linear function and auxiliary variables.

    This function simplify duplicates and linearizes multi-site cluster terms.

    Args:
        cluster_terms(list of tuples of (list of int, float)):
            A list of tuples, each represents a cluster term in the energy
            representation, containing indices of variables to be taken product with,
            and the factor before the boolean product.
            Energy is taken per super-cell.
        variables(cp.Variable):
            cvxpy variables storing the ground-state result.
        coefficients_cutoff(float): optional
            Minimum cutoff to the coefficient of terms. If the absolute value of a
            term coefficient is less than cutoff, it will not be included in the
            objective function. If no cutoff is given, will include every term.
    Returns:
        cp.Expression, cp.Variable, list of lists of int, list of Constraint:
            The linearized energy expression, auxiliary slack variables for each
            multi-body product term, a list containing the indices of variables whose
            product equals to the corresponding auxiliary slack variable, and
            linearize constraints for each multi-body product term.
    """
    # Simplify cluster terms first.
    sorted_terms = [
        (tuple(sorted(inds)), fac1, fac2) for inds, fac1, fac2 in cluster_terms
    ]
    simplified_terms = {}
    for inds, fac1, fac2 in sorted_terms:
        if inds not in simplified_terms:
            simplified_terms[inds] = fac1 * fac2
        else:
            simplified_terms[inds] += fac1 * fac2
    simplified_terms = {
        inds: coef
        for inds, coef in simplified_terms.items()
        if abs(coef) >= coefficients_cutoff
    }
    simplified_terms = sorted(simplified_terms.items(), key=lambda t: (len(t[0]), t[1]))

    n_const = len([inds for inds, _ in simplified_terms if len(inds) == 0])
    n_single = len([inds for inds, _ in simplified_terms if len(inds) == 1])
    n_slack = len([inds for inds, _ in simplified_terms if len(inds) > 1])
    if n_slack == 0:
        aux_variables = None
    else:
        aux_variables = cp.Variable(n_slack, boolean=True)
    aux_constraints = []
    # Vectorized objective functions to speed up compilation.
    # Only one constant term is possible at most.
    expression = simplified_terms[0][1] if n_const == 1 else 0
    # Summing point terms.
    point_inds = [
        inds[0] for inds, _ in sorted(simplified_terms[n_const : n_const + n_single])
    ]
    point_coefs = np.array(
        [c for _, c in sorted(simplified_terms[n_const : n_const + n_single])]
    )
    if n_single > 0:
        expression += variables[point_inds] @ point_coefs
    # Summing many-body terms.
    many_inds = [
        list(inds)
        for inds, _ in simplified_terms[
            n_const + n_single : n_const + n_single + n_slack
        ]
    ]
    many_coefs = np.array(
        [
            coef
            for _, coef in simplified_terms[
                n_const + n_single : n_const + n_single + n_slack
            ]
        ]
    )
    if n_slack > 0:
        expression += aux_variables @ many_coefs
    for aux_id, inds in enumerate(many_inds):
        for var_id in inds:
            aux_constraints.append(aux_variables[aux_id] <= variables[var_id])
        aux_constraints.append(
            aux_variables[aux_id] >= cp.sum(variables[inds]) + 1 - len(inds)
        )

    if not isinstance(expression, cp.Expression):
        raise RuntimeError(
            f"The energy function {expression} has no configuration"
            f" degree of freedom. Cannot be optimized!"
        )

    return expression, aux_variables, many_inds, aux_constraints



"""Test utility functions for upper-bound."""

from itertools import product

import cvxpy as cp
import numpy as np
import numpy.testing as npt
import pytest
from pymatgen.core import Species

from smol.capp.generate.groundstate.upper_bound.indices import (
    get_sublattice_indices_by_site,
    get_variable_indices_for_each_composition_component,
    map_ewald_indices_to_variable_indices,
)
from smol.capp.generate.groundstate.upper_bound.terms import (
    get_auxiliary_variable_values,
    get_expression_and_auxiliary_from_terms,
)
from smol.capp.generate.groundstate.upper_bound.variables import (
    get_variables_from_sublattices,
)
from smol.cofe.space.domain import Vacancy, get_allowed_species


def test_variable_indices_for_components(
    solver_test_ensemble, solver_test_initial_occupancy
):
    _, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )

    var_inds_for_components = get_variable_indices_for_each_composition_component(
        solver_test_ensemble.sublattices,
        variable_indices,
        solver_test_ensemble.processor.structure,
    )
    # Total 7 species on all sub-lattices.
    assert len(var_inds_for_components) == 7
    dim_id = 0
    for sublattice in solver_test_ensemble.sublattices:
        sub_bits = sublattice.species
        sl_active_variables = np.array(
            [
                [v for v in variable_indices[i] if v >= 0]
                for i in range(solver_test_ensemble.num_sites)
                if i in sublattice.active_sites
            ],
            dtype=int,
        )
        if Species("Li", 1) in sub_bits:
            # 1 li sites are restricted.
            li_code = sublattice.encoding[sub_bits.index(Species("Li", 1))]
            va_code = sublattice.encoding[sub_bits.index(Vacancy())]
            restricted_all_sites = sublattice.restricted_sites
            assert len(restricted_all_sites) == 3
            restricted_li_sites = restricted_all_sites[
                np.where(
                    solver_test_initial_occupancy[restricted_all_sites] == li_code
                )[0]
            ]
            restricted_vac_sites = np.setdiff1d(
                restricted_all_sites, restricted_li_sites
            )
            assert len(restricted_li_sites) == 1
            assert len(restricted_vac_sites) == 2
            # Only li sites restricted.
            npt.assert_array_equal(
                solver_test_initial_occupancy[restricted_li_sites], li_code
            )
            npt.assert_array_equal(
                solver_test_initial_occupancy[restricted_vac_sites], va_code
            )
            for sp_id, species in enumerate(sub_bits):
                var_ids, n_fix = var_inds_for_components[dim_id]
                if species == Species("Li", 1):
                    assert n_fix == 1  # 1 li sites manually restricted
                elif species == Vacancy():
                    assert n_fix == 2
                else:
                    raise ValueError(
                        "Li/Vac sub-lattice was not correctly partitioned!"
                        f" Extra species {species}."
                    )
                # 2 unrestricted li sites + 1 unrestricted vac sites.
                assert len(var_ids) == 3
                # Check indices are correct.
                npt.assert_array_equal(sl_active_variables[:, sp_id], var_ids)
                dim_id += 1
        elif Species("Mn", 2) in sub_bits:
            # No restricted site, all 6 sites should be active.
            for sp_id in range(len(sub_bits)):
                var_ids, n_fix = var_inds_for_components[dim_id]
                assert n_fix == 0
                assert len(var_ids) == 6
                npt.assert_array_equal(sl_active_variables[:, sp_id], var_ids)
                dim_id += 1
        elif Species("O", -2) in sub_bits:  # O2- sublattice totally inactive.
            assert list(sub_bits) == [Species("O", -2)]
            var_ids, n_fix = var_inds_for_components[dim_id]
            assert n_fix == 9
            assert len(var_ids) == 0
            dim_id += 1
        else:  # F sub-lattice totally inactive.
            assert list(sub_bits) == [Species("F", -1)]
            var_ids, n_fix = var_inds_for_components[dim_id]
            assert n_fix == 3
            assert len(var_ids) == 0
            dim_id += 1


def test_ewald_indices(solver_test_ensemble, solver_test_initial_occupancy):
    _, variable_indices = get_variables_from_sublattices(
        solver_test_ensemble.sublattices,
        solver_test_ensemble.processor.structure,
        solver_test_initial_occupancy,
    )
    assert len(variable_indices) == solver_test_ensemble.num_sites
    ew_processor = solver_test_ensemble.processor.processors[-1]
    site_sublattice_ids = get_sublattice_indices_by_site(
        solver_test_ensemble.sublattices
    )

    ew_to_var_id = map_ewald_indices_to_variable_indices(
        ew_processor.structure,
        variable_indices,
    )

    n_ew_rows = len(solver_test_ensemble.processor.processors[-1]._ewald_structure)
    site_spaces = get_allowed_species(solver_test_ensemble.processor.structure)
    assert len(ew_to_var_id) == n_ew_rows
    ew_id = 0
    var_id = 0

    restricted_vac_sites = solver_test_ensemble.restricted_sites[
        np.where(
            solver_test_initial_occupancy[solver_test_ensemble.restricted_sites] == 4
        )[0]
    ]
    # print("supercell:\n", ew_processor.structure)
    # print("Initial occu:", solver_test_initial_occupancy)
    # print("sub-lattices:", solver_test_ensemble.sublattices)
    # print("site sub-lattice indices:", site_sublattice_ids)
    # print("variable indices:", variable_indices)
    # print("restricted_sites:", solver_test_ensemble.restricted_sites)
    # print("restricted_vac_sites:", restricted_vac_sites)
    # print("ew_to_var_id:\n", ew_to_var_id)

    # This test is only applied to the solver_test ensemble in conftest.
    expects = []
    for site_id, site_space in enumerate(site_spaces):
        sublattice_id = site_sublattice_ids[site_id]
        sublattice = solver_test_ensemble.sublattices[sublattice_id]
        for spec in site_space:
            if isinstance(spec, Vacancy):
                # In the Li/Vac sub-lattice.
                if Species("Li", 1) in sublattice.species:
                    # Vacancy in an active site also occupies a variable.
                    # Vacancy in an inactive site does not.
                    if site_id in sublattice.active_sites:
                        var_id += 1
                continue
            if site_id in solver_test_ensemble.restricted_sites:
                # Not the inactive F or O sub-lattice, just manually restricted.
                if (Species("F", -1) not in sublattice.species) and (
                    Species("O", -2) not in sublattice.species
                ):
                    # Always occupied by one non-vacancy species.
                    if spec == Species("Li", 1) and site_id not in restricted_vac_sites:
                        expected = -1
                    # Always occupied by other species than this one.
                    else:
                        expected = -2
                # Inactive F or O sub-lattice.
                else:
                    if Species("F", -1) in sublattice.species:
                        if spec == Species("F", -1):
                            expected = -1
                        else:
                            expected = -2
                    elif Species("O", -2) in sublattice.species:
                        if spec == Species("O", -2):
                            expected = -1
                        else:
                            expected = -2
                    else:
                        raise ValueError(f"Invalid sublattice: {sublattice}!")
            # Active site.
            else:
                # In the new sub-lattice.
                if spec in sublattice.species:
                    expected = var_id
                    var_id += 1
                # Not in the new sub-lattice.
                else:
                    expected = -2

            ew_id += 1
            expects.append(expected)

    # print("expected:\n", expects)
    assert ew_id == n_ew_rows
    npt.assert_array_equal(expects, ew_to_var_id)


def test_expression_from_terms():
    x = cp.Variable(2, boolean=True)
    # Empty expression.
    terms = []
    with pytest.raises(RuntimeError):
        _ = get_expression_and_auxiliary_from_terms(terms, x)
    # Expression with only constant.
    terms = [([], -1, 1.0), ([], 100, 1.0)]
    with pytest.raises(RuntimeError):
        _ = get_expression_and_auxiliary_from_terms(terms, x)
    # Correct expression, which is to be evaluated.
    # A very simple test case.
    # 0 + 2 x0 -3 x1 + x0 * x1.
    terms = [
        ([], -1, 1),
        ([], 1, 1),
        ([0], 2, 1),
        ([1], -3, 1),
        ([0, 1], -1, 1),
        ([1, 0], 2, 1),
    ]
    func, y, indices, aux_cons = get_expression_and_auxiliary_from_terms(terms, x)
    assert y.size == 1
    assert len(indices) == 1
    assert list(indices[0]) == [0, 1]
    assert len(aux_cons) == 3
    assert isinstance(func, cp.Expression)
    for val0, val1 in product(*[[0, 1], [0, 1]]):
        true_val = 2 * val0 - 3 * val1 + val0 * val1
        x.value = np.array([val0, val1], dtype=int)
        y.value = [val0 * val1]
        assert func.value == true_val
        # Auxiliary constraints should always be satisfied.
        for con in aux_cons:
            assert con.value()

    for rand_vals in product(*[[0, 1], [0, 1]]):
        rand_vals = np.array(rand_vals)
        rand_aux_vals = get_auxiliary_variable_values(rand_vals, indices)
        for ii, inds in enumerate(indices):
            assert len(inds) > 1
            assert np.prod(rand_vals[inds]) == rand_aux_vals[ii]



"""Tools to characterize MC convergence."""

__author__ = "Ronald Kam"

from warnings import warn

import numpy as np


def check_property_converged(
    property_array, conv_tol=None, min_std=1e-4, last_m=None, verbose=False
):
    """Check if a property is converged in MC run.

    A relatively rudimentary check for convergence, perhaps not the most rigorous.
    The criteria for convergence are:
     1) The last value of the property should be close to the mean of the property, +/-
        some tolerance.

     2) The cumulative mean of the property in the last M samples should be close to
        the mean of the property, +/- some tolerance.

    Args:
        property_array (ndarray):
            array of numerical values of a property (energy, composition, etc)
        conv_tol (float):
            value to determine the convergence tolerance. If None, the std. dev. of the
            property will be used
        min_std (float):
            minimum standard deviation of the property required to perform this
            convergence test. If std dev is too small, then we can assume that there
            were few acceptances and that the simulation is converged.
        last_m (int):
            Number of last M samples to determine the convergence of cumulative property
            mean. If None, take the last 10% of samples.

    Returns:
        converged (bool):
            True if converged, False if not

    """
    property_array = np.array(property_array)
    std_prop = np.std(property_array)
    if std_prop < min_std:
        # Check will never pass if the property std dev is too small, which means there
        # were very few acceptances. Assume in this case that we are already close to
        # the equilibrium value and that MC is converged. This is common for MC runs at
        # low temperature.

        if verbose:
            print(
                "The std dev of the property is very small, so it can be assumed that"
                "MC is converged."
            )
        return True

    if conv_tol is None:
        conv_tol = std_prop

    mean_prop = np.average(property_array)
    n_samples = len(property_array)

    if last_m is None:
        last_m = int(n_samples / 10)

    elif last_m > n_samples:
        warn(
            f"The specified number of last M samples ({last_m}) is greater than the "
            f"number of samples ({n_samples})! Changing to the default value of last "
            f"10% of samples ({int(n_samples / 10)})"
        )
        last_m = int(n_samples / 10)

    # Check criteria 1
    converged_last_val = 0
    if (
        property_array[-1] < mean_prop + conv_tol
        and property_array[-1] > mean_prop - conv_tol
    ):
        converged_last_val = True

    else:
        if verbose:
            print("The last value of the property is not close to the mean.")

    # Check criteria 2
    converged_cum_mean = 0
    prop_cum_mean = np.divide(np.cumsum(property_array), np.arange(1, n_samples + 1))
    if all(prop_cum_mean[-last_m:] < mean_prop + conv_tol) and all(
        prop_cum_mean[-last_m:] > mean_prop - conv_tol
    ):
        converged_cum_mean = True

    else:
        if verbose:
            print("The cumulative property mean does not converge to the global mean.")

    if converged_last_val and converged_cum_mean:
        return True

    else:
        return False


def determine_discard_number(
    property_array, init_discard=None, increment=10, verbose=False
):
    """Determine the minimum number of samples to discard for MC equilibration.

    Identify a range of number of samples to discard. Determine the smallest value that
    yields a converged property.

    Args:
        property_array (ndarray):
            array of numerical values of a property (energy, composition, etc)
        init_discard (int):
            initial guess for number of samples to discard. If None, take the first 10%
            as the initial guess
        increment (int):
            determines the increment of discard values to search for. Will search for
            discard values in increments of (n_samples - init_discard)/increment

    Returns:
        discard_n (int):
            number of samples to discard

    """
    property_array = np.array(property_array)
    n_samples = len(property_array)

    if init_discard is None:
        init_discard = int(n_samples / 10)
    elif init_discard > n_samples:
        warn(
            f"Number to discard ({init_discard}) is greater than num samples "
            f"({n_samples})! Will set the discard value to be 10% of the number of"
            f"samples ({int(n_samples/10)})"
        )
        init_discard = int(n_samples / 10)

    discard_val_increment = (n_samples - init_discard) / increment
    for discard_n in np.arange(
        init_discard, n_samples - discard_val_increment, discard_val_increment
    ):
        discard_n = int(discard_n)
        if check_property_converged(property_array[discard_n:]):
            return discard_n

    if verbose:
        print(
            "Could not find a discard value that leads to MC convergence. Will return "
            "0."
        )

    return 0



"""Implementations of general descriptors to manage attributes."""

from collections.abc import Sequence


class SetMany:
    """Descriptor to set an attributed value that cascades to a sequence of objects.

    This descriptor is useful when an attribute is needs to be synchronized between
    a class that holds a sequence of objects that have a corresponding attribute.
    """

    def __init__(self, attr_name, container_name, container_attr_name=None):
        """Define the name of the attribute and the container.

        Args:
            attr_name (str):
                name of the attribute to set.
            container_name (str):
                name of the container attribute holding instances with attributes to
                be set to the same value as the attribute being set.
            container_attr_name (str): optional
                name of the attribute in the container that holds the instances.
        """
        self._attr_name = attr_name
        self._container_name = container_name
        self._container_attr_name = (
            container_attr_name if container_attr_name else attr_name
        )

    def __get__(self, instance, objtype=None):
        """Get the value of the attribute."""
        # we don't need to get all values, and maybe we should check they are all the
        # same
        values = [
            getattr(obj, self._container_attr_name)
            for obj in self._container_objs(instance)
        ]
        return values[0]

    def __set__(self, instance, value):
        """Set the value of the attribute and cascade to the container."""
        for obj in self._container_objs(instance):
            setattr(obj, self._container_attr_name, value)

    def _container_objs(self, instance):
        """Yield the values in the container."""
        container = getattr(instance, self._container_name)
        if isinstance(container, Sequence):
            iterable = container
        elif hasattr(container, "values"):
            iterable = container.values()
        else:
            raise TypeError(
                "container must be an iterable or a dict-like object, not {}".format(
                    type(container)
                )
            )

        yield from iterable



import numpy as np
import pytest

from smol.capp.generate import generate_random_ordered_occupancy
from smol.cofe import ClusterSubspace
from smol.moca import ClusterExpansionProcessor


@pytest.mark.parametrize(
    "composition, charge_neutral", [(True, False), (None, True), (None, False)]
)
def test_generate_random_ordered_occupancy(
    single_structure, composition, charge_neutral
):
    subspace = ClusterSubspace.from_cutoffs(
        single_structure, cutoffs={}, supercell_size="volume"
    )
    scm = np.eye(3)
    scm[0, 0] = 5
    scm[1, 1] = 4
    processor = ClusterExpansionProcessor(subspace, scm, np.ones(len(subspace)))
    print(processor.num_sites)

    if composition is True:
        composition = [sl.composition for sl in processor.get_sublattices()]
        print([sl.sites.shape for sl in processor.get_sublattices()])
        print(composition)

    for _ in range(5):
        occu = generate_random_ordered_occupancy(
            processor, composition=composition, charge_neutral=charge_neutral
        )
        structure_ = processor.structure_from_occupancy(occu)
        assert structure_.is_ordered

        if composition is not None:
            assert (
                structure_.composition.fractional_composition
                == single_structure.composition.fractional_composition
            )

        if charge_neutral is True:
            assert structure_.charge == 0
